# -*- coding: utf-8 -*-
'''
environments module for reinforcement learning
==============================================

This module contains classes that provides tools to load/save objects and
environments, add/remove `agents`/`subjects`, assign `agents` to `subjects` and
run models.

Classes
-------
Environment:
    The base class of all environment classes.

EnvironmentStaticMap:
    An environment with static interaction map.
'''

from .environment import Environment  # noqa: W0611
from .environment_static_map import EnvironmentStaticMap  # noqa: W0611



# -*- coding: utf-8 -*-
'''
EndlessBuffer class
===================

A `Buffer` without size limit.
'''

from typing import Dict, List, Optional, cast

from reil.datatypes.buffers import Buffer, PickModes, T


class EndlessBuffer(Buffer[T]):
    '''
    A `Buffer` without size limit.

    Extends `Buffer` class.
    '''

    def __init__(self,
                 buffer_names: Optional[List[str]] = None,
                 pick_mode: Optional[PickModes] = None) -> None:
        '''
        Arguments
        ---------
        buffer_names:
            A list containing the names of buffer queues.

        pick_mode:
            The default mode to pick items from the list.
        '''
        self.setup(buffer_names, pick_mode)

    def setup(self,
              buffer_names: Optional[List[str]] = None,
              pick_mode: Optional[PickModes] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_names:
            A list containing the names of buffer elements.

        pick_mode:
            The default mode to pick items from the list.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        super().setup(buffer_names=buffer_names, pick_mode=pick_mode)

    def add(self, data: Dict[str, T]) -> None:
        '''
        Add a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.
        '''
        self._buffer_index += 1
        for key, v in data.items():
            self._buffer[key].append(v)
        self._count += 1

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        super().reset()
        if self._buffer_names is not None:
            self._buffer = cast(Dict[str, List[T]],
                                dict((name, [])
                                     for name in self._buffer_names))
        else:
            self._buffer = None



# -*- coding: utf-8 -*-
'''
CircularBuffer class
====================

A `Buffer` that overflows!
'''

from typing import Dict, Tuple

from reil.datatypes.buffers import Buffer, T


class CircularBuffer(Buffer[T]):
    '''
    A `Buffer` that overflows.

    Extends `Buffer` class.
    '''
    _buffer_full = False

    def add(self, data: Dict[str, T]) -> None:
        '''
        Add a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.

        Notes
        -----
        If the buffer is full, new items will be writen over the oldest one.
        '''
        try:
            super().add(data)
        except IndexError:
            self._buffer_full = True
            self._buffer_index = -1
            super().add(data)

        # the size does not change if buffer is full.
        self._count -= self._buffer_full

    def _pick_old(self, count: int) -> Dict[str, Tuple[T, ...]]:
        '''
        Return the oldest items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.
        '''
        if self._buffer_full:
            slice_pre = slice(self._buffer_index + 1,
                              self._buffer_index + count + 1)
            slice_post = slice(
                max(0, count - (self._buffer_size - self._buffer_index) + 1))
            return dict((name, tuple(
                buffer[slice_pre] +
                buffer[slice_post]))
                for name, buffer in self._buffer.items())
        else:
            return super()._pick_old(count)

    def _pick_recent(self, count: int) -> Dict[str, Tuple[T, ...]]:
        '''
        Return the most recent items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.
        '''
        if count - self._buffer_index <= 1 or not self._buffer_full:
            return super()._pick_recent(count)
        else:
            slice_pre = slice(-(count - self._buffer_index - 1), None)
            slice_post = slice(self._buffer_index + 1)
            return dict((name, tuple(
                buffer[slice_pre] +
                buffer[slice_post]))
                for name, buffer in self._buffer.items())

    def _pick_all(self) -> Dict[str, Tuple[T, ...]]:
        '''
        Return all items in the buffer.
        '''
        if self._buffer_full:
            slice_pre = slice(self._buffer_index + 1, None)
            slice_post = slice(self._buffer_index + 1)
            return dict((name, tuple(
                         buffer[slice_pre] +
                         buffer[slice_post]))
                        for name, buffer in self._buffer.items())
        else:
            return super()._pick_all()

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        super().reset()
        self._buffer_full = False



# -*- coding: utf-8 -*-
'''
Learner class
=============

The base class for all `learner` classes.
'''
import pathlib
from typing import Any, Generic, Optional, Tuple, TypeVar, Union

from reil import learners, reilbase
from reil.datatypes import ReilData

LabelType = TypeVar('LabelType')


class Learner(reilbase.ReilBase, Generic[LabelType]):
    '''
    The base class for all `learner` classes.
    '''
    def __init__(self,
                 learning_rate: learners.LearningRateScheduler,
                 **kwargs: Any) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the learning rate
            based on epoch. If any scheduler other than constant is provided,
            the model uses the `new_rate` method of the scheduler to determine
            the learning rate at each epoch.
        '''
        super().__init__(**kwargs)
        self._learning_rate = learning_rate

    @classmethod
    def from_pickle(cls,
                    filename: str,
                    path: Optional[Union[pathlib.Path, str]] = None):
        instance = cls(learning_rate=learners.ConstantLearningRate(0.0))
        instance.load(filename=filename, path=path)

        return instance

    def predict(self, X: Tuple[ReilData, ...]) -> Tuple[LabelType, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `ReilData` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        raise NotImplementedError

    def learn(self, X: Tuple[ReilData, ...], Y: Tuple[LabelType, ...]) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `ReilData` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        raise NotImplementedError



# -*- coding: utf-8 -*-
'''
ReilData class
==============

The main datatype used to communicate `state`s, `action`s, and `reward`s,
between objects in `reil`. `ReilData` is basically a tuple that contains
instances of `BaseData`, `CategoricalData`, and `NumericalData`.
'''
from __future__ import annotations

import dataclasses
import itertools
from dataclasses import field
from typing import (Any, Callable, Dict, Generic, Iterable, Iterator, List,
                    Mapping, Optional, Sequence, Tuple, TypeVar, Union, cast,
                    overload)

from typing_extensions import Literal

T = TypeVar('T')

Categorical = TypeVar('Categorical')
Numerical = TypeVar('Numerical', int, float)

Normal = Union[Literal[0], Literal[1], float]
Normalized = Union[Normal, Sequence[Normal], None]
Normalizer = Callable[..., Normalized]


@dataclasses.dataclass(frozen=True)
class BaseData(Generic[T]):
    '''
    The base class for `ReilData` elements.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).


    :meta private:
    '''
    name: str
    value: Optional[Union[T, Sequence[T]]] = None
    normalizer: Optional[Normalizer] = lambda _: None
    lazy_evaluation: Optional[bool] = field(default=None, compare=False)
    is_numerical: Optional[bool] = field(
        default=None, init=False, repr=False, compare=False)
    _normalized: Normalized = field(
        default=None, init=False, repr=False, compare=False)

    def __post_init__(self):
        # if not isinstance(self.value, Hashable):
        #     warnings.warn('Non-hashable object!')
        self._validate(self)
        if self.normalizer is None:
            object.__setattr__(self, 'normalizer',
                               self._default_normalizer)
        if not self.lazy_evaluation:
            object.__setattr__(self, '_normalized',
                               self.normalizer(self))

    @staticmethod
    def _validate(data: BaseData[T]) -> None:
        '''
        Validate the input data.

        Arguments
        ---------
        data:
            An instance to be validated.

        Raises
        ------
        ValueError:
            if `data.value` does not match the validation criteria.


        :meta public:
        '''
        pass

    @staticmethod
    def _default_normalizer(data: BaseData[T]) -> Normalized:
        '''
        Normalize the data.

        Arguments
        ---------
        data:
            An instance to be normalized.

        Returns
        -------
        :
            The normalized form of `data.value`.
        '''
        return None

    @property
    def normalized(self) -> Normalized:
        if self._normalized is None:
            object.__setattr__(self, '_normalized',
                               self.normalizer(self))

        return self._normalized

    def as_dict(self) -> Dict[str, Any]:
        '''
        Return the data as a dictionary.

        Returns
        -------
        :
            The data as a dictionary.
        '''
        return {'name': self.name, 'value': self.value}


@dataclasses.dataclass(frozen=True)
class CategoricalData(BaseData[Categorical]):
    '''
    A datatype for categorical data.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).

    categories:
        A list of all categories.
    '''
    categories: Optional[Tuple[Categorical, ...]] = None

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'is_numerical', False)

    @staticmethod
    def _default_normalizer(data: CategoricalData) -> Normalized:
        value = data.value
        categories = data.categories
        if categories is None or value is None:
            return None

        if isinstance(value, type(categories[0])):
            return tuple(int(x_i == value)
                         for x_i in categories)

        return tuple(int(x_i == v)
                     for v in value
                     for x_i in categories)

    @staticmethod
    def _validate(data: CategoricalData) -> None:
        value = data.value
        if value is not None:
            is_sequence = isinstance(value, (list, tuple))

            categories = data.categories
            if categories is not None:
                if ((not is_sequence and value not in categories) or
                        (is_sequence and any(v not in categories
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is '
                        f'in the categories={categories}.')

    def as_dict(self) -> Dict[str, Any]:
        return {'name': self.name, 'value': self.value,
                'categories': self.categories,
                'is_numerical': self.is_numerical}


@dataclasses.dataclass(frozen=True)
class NumericalData(BaseData[Numerical]):
    '''
    A datatype for numerical data.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).

    lower:
        The lower bound of value.

    upper:
        The upper bound of value.
    '''
    lower: Optional[Numerical] = None
    upper: Optional[Numerical] = None

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'is_numerical', True)

    @staticmethod
    def _default_normalizer(data: NumericalData) -> Normalized:
        value = data.value
        lower = data.lower
        upper = data.upper
        if value is None or lower is None or upper is None:
            return None

        denominator = upper - lower

        sequence_check = isinstance(value, (list, tuple))
        try:
            if sequence_check:
                return tuple((v - lower) / denominator
                             for v in value)
            else:
                return (value - lower) / denominator
        except ZeroDivisionError:
            return [1] * len(value) if sequence_check else 1

    @staticmethod
    def _validate(data: NumericalData) -> None:
        value = data.value
        lower = data.lower
        upper = data.upper
        if value is not None:
            is_sequence = isinstance(value, (list, tuple))

            if lower is not None:
                if ((not is_sequence and value < lower) or
                        (is_sequence and any(v < lower
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is less than lower={lower}.')

            if upper is not None:
                if ((not is_sequence and value > upper) or
                        (is_sequence and any(v > upper
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is '
                        f'greater than upper={upper}.')

    def as_dict(self) -> Dict[str, Any]:
        return {'name': self.name, 'value': self.value,
                'lower': self.lower, 'upper': self.upper,
                'is_numerical': self.is_numerical}


ReilDataInput = Union[Mapping[str, Any],
                      BaseData[Any],
                      CategoricalData[Categorical],
                      NumericalData[Numerical]]


class ReilData(Sequence[ReilDataInput], Generic[Categorical, Numerical]):
    '''
    The main datatype used to communicate `state`s, `action`s, and `reward`s,
    between objects in `reil`.
    '''

    def __init__(self,
                 data: Union[ReilDataInput,
                             Sequence[ReilDataInput],
                             Iterator[ReilDataInput]],
                 lazy_evaluation: Optional[bool] = None):
        '''
        Arguments
        ---------
        data:
            One or a sequence of `BaseData`, `CategoricalData`,
            `NumericalData`, or dicts that include 'name'. Other attributes are
            optional. If none of `categories`, `lower` and `upper` are
            provided, the object is assumed `BaseData`.

        lazy_evaluation:
            Whether to compute normalized value at the time of instantiation or
            at the time of the first access. If not provided, `ReilData` looks
            for `lazy_evaluation` in each item. If failed, True is assumed.
        '''
        temp: List[Union[BaseData[Any],
                         CategoricalData[Categorical],
                         NumericalData[Numerical]]] = []
        _data = data if isinstance(data, (Sequence, Iterator)) else [data]
        for d in _data:
            if isinstance(d, BaseData):
                temp.append(d)
            elif isinstance(d, dict):
                name = d['name']
                value = d.get('value')
                normalizer = d.get('normalizer')
                l_e = lazy_evaluation if lazy_evaluation is not None \
                    else d.get('lazy_evaluation', True)

                if 'categories' in d:
                    temp.append(CategoricalData(
                        name=name,
                        value=value,
                        categories=d['categories'],
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
                elif 'lower' in d or 'upper' in d:
                    temp.append(NumericalData(
                        name=name,
                        value=value,
                        lower=d.get('lower'),
                        upper=d.get('upper'),
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
                else:
                    temp.append(BaseData(
                        name=name,
                        value=value,
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
            else:
                raise TypeError(f'Unknow input type {type(d)} for item: {d}')

        self._data = tuple(temp)
        self._clear_temps()

    @classmethod
    def single_base(
            cls,
            name: str,
            value: Optional[Union[T, Sequence[T]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            BaseData(name=name,
                     value=value,
                     normalizer=normalizer,
                     lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_categorical(
            cls,
            name: str,
            value: Optional[Union[Categorical, Sequence[Categorical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            categories: Optional[Tuple[Categorical, ...]] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        categories:
            A list of categories if the object is categorical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            CategoricalData(
                name=name,
                value=value,
                categories=categories,
                normalizer=normalizer,
                lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_numerical(
            cls,
            name: str,
            value: Optional[Union[Numerical, Sequence[Numerical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            lower: Optional[Numerical] = None,
            upper: Optional[Numerical] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        lower:
            The lower bound of value if data is numerical

        upper:
            The upper bound of the value is numerical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            NumericalData(
                name=name,
                value=value,
                lower=lower,
                upper=upper,
                normalizer=normalizer,
                lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_item(
            cls,
            name: str,
            value: Optional[Union[
                Numerical, Sequence[Numerical],
                Categorical, Sequence[Categorical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            categories: Optional[Tuple[Categorical, ...]] = None,
            lower: Optional[Numerical] = None,
            upper: Optional[Numerical] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        categories:
            A list of categories if the object is categorical

        lower:
            The lower bound of value if data is numerical

        upper:
            The upper bound of the value is numerical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        if lower is not None or upper is not None:
            instance._data = (
                NumericalData(
                    name=name,
                    value=value,  # type: ignore
                    lower=lower,
                    upper=upper,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)
        elif categories is not None:
            instance._data = (
                CategoricalData(
                    name=name,
                    value=value,
                    categories=categories,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)
        else:
            instance._data = (
                CategoricalData(
                    name=name,
                    value=value,
                    categories=categories,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)

        return instance

    def _clear_temps(self):
        self._value: Optional[Dict[str, Any]] = None
        self._lower: Optional[Dict[str, Numerical]] = None
        self._upper: Optional[Dict[str, Numerical]] = None
        self._categories: Optional[Dict[str, Tuple[Categorical, ...]]] = None
        self._is_numerical: Optional[Dict[str, bool]] = None

    def index(self, value: Any,
              start: int = 0, stop: Optional[int] = None) -> int:
        _stop = stop if stop is not None else len(self._data)
        if isinstance(value, BaseData):
            for i in range(start, _stop):
                if self._data[i] == value:
                    return i

            raise ValueError(f'{value} is not on the list.')

        elif isinstance(value, type(self._data[0].name)):
            for i in range(start, _stop):
                if self._data[i].name == value:
                    return i

            raise ValueError(f'{value} is not on the list.')

        else:
            raise ValueError(f'{value} is not on the list.')

    @property
    def value(self) -> Dict[str, Any]:
        '''
        Return a dictionary with elements' names as keys and
        their respective values as values.

        Returns
        -------
        :
            Names of the elements and their values.
        '''
        if self._value is None:
            self._value = dict((v.name, v.value)
                               for v in self._data)

        return self._value

    @property
    def lower(self) -> Dict[str, Numerical]:
        '''
        Return all `lower` attributes.

        Returns
        -------
        :
            `lower` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._lower is None:
            self._lower = dict(
                (v.name, cast(Numerical, v.lower))  # type: ignore
                for v in self._data
                if hasattr(v, 'lower'))

        return self._lower

    @property
    def upper(self) -> Dict[str, Numerical]:
        '''
        Return all `upper` attributes.

        Returns
        -------
        :
            `upper` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._upper is None:
            self._upper = dict(
                (v.name, cast(Numerical, v.upper))  # type: ignore
                for v in self._data
                if hasattr(v, 'upper'))

        return self._upper

    @property
    def categories(self) -> Dict[str, Tuple[Categorical, ...]]:
        '''
        Return all `categories` attributes.

        Returns
        -------
        :
            `categories` attribute of all `CategoricalData` variables with
            their names as keys.
        '''
        if self._categories is None:
            self._categories = dict(
                (v.name, cast(Categorical, v.categories))  # type: ignore
                for v in self._data
                if hasattr(v, 'categories'))

        return self._categories

    @property
    def normalized(self) -> ReilData:
        '''
        Normalize all items in the instance.

        Returns
        -------
        :
            A `ReilData` consist of normalized values of all the items in the
            instance, in the form of `NumericalData` objects.
        '''
        return ReilData(
            NumericalData(name=v.name, value=v.normalized,
                          lower=0, upper=1, lazy_evaluation=True)
            for v in self._data)

    def flatten(self) -> List[BaseData]:
        """Combine values of all items in the instance.

        Returns
        -------
        :
            A list that contains all the values of all the items.
        """
        def make_iterable(x: T) -> Iterable[T]:
            return x if isinstance(x, Iterable) else [x]

        return list(itertools.chain(*[make_iterable(sublist)
                                      for sublist in self.value.values()]))

    def split(self) -> Union[ReilData, List[ReilData]]:
        """Split the `ReilData` into a list of `ReilData`.

        Returns
        -------
        :
            All items in the instance as separate `ReilData` instances.
        """
        if len(self) == 1:
            d = self._data[0]
            if not isinstance(d.value, (list, tuple)):
                splitted_list = ReilData(self._data)
            else:
                temp = d.as_dict()
                cls = type(d)
                value = temp['value']
                del temp['value']
                if 'is_numerical' in temp:
                    del temp['is_numerical']

                splitted_list = [
                    ReilData(cls(
                        value=v,
                        **temp))
                    for v in value]

        else:
            splitted_list = cast(List[ReilData], list(self._data))

        return splitted_list

    @overload
    def __getitem__(self, i: int) -> BaseData:
        ...

    @overload
    def __getitem__(self, s: slice) -> Tuple[BaseData, ...]:
        ...

    def __getitem__(self, x):
        return self._data.__getitem__(x)

    def __len__(self) -> int:
        return self._data.__len__()

    def __hash__(self) -> int:
        return hash(tuple(self._data))

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self)) and (self._data == other._data)

    def __add__(self, other: Union[BaseData[T], ReilData]) -> ReilData:
        if isinstance(other, BaseData):
            if other.name in self._data:
                raise ValueError(
                    'Cannot have items with same names.'
                    ' Use update() if you need to update an item.')
            new_data = [other]
        elif isinstance(other, ReilData):
            for k in other:
                if k in self._data:
                    raise ValueError(
                        'Cannot have items with same names.'
                        ' Use update() if you need to update an item.')
            new_data = other._data
        else:
            raise TypeError(
                'Concatenation of type ReilData'
                f' and {type(other)} not implemented!')

        return ReilData(itertools.chain(self._data, new_data))

    def __neg__(self) -> ReilData:
        temp = [v.as_dict()
                for v in self._data]
        for item in temp:
            item['value'] = -item['value']

        return ReilData(temp)

    def __repr__(self) -> str:
        return f'[{super().__repr__()} -> {self._data}]'

    def __str__(self) -> str:
        return f"[{', '.join((d.__str__() for d in self._data))}]"


if __name__ == "__main__":
    # from timeit import timeit

    def f1():
        t1 = ReilData.single_categorical(name='test A', value=['a', 'b'],
                                         categories=('a', 'b'))
        t1 = ReilData.single_numerical(name='test A', value=[1, 2])
        t1 = ReilData.single_base(name='test A', value=['a', 'b'])
        return t1

    def f2():
        t2 = ReilData(({'name': x, 'value': x,
                        'categories': list('abcdefghijklmnopqrstuvwxyz')}
                       for x in 'abcdefghijklmnopqrstuvwxyz'),
                      lazy_evaluation=True)
        return t2.normalized.flatten()

    def f3():
        t2 = ReilData([{'name': 'A', 'value': 'a', 'categories': ['a', 'b']},
                       {'name': 'B', 'value': [10, 20],
                        'lower': 0, 'upper': 100}], lazy_evaluation=True)
        return t2

    # t1 = f1()
    # t2 = ReilData(t1)
    print(f1() + f3())
    # test = f1().split()
    # print(test)
    # print(timeit(f1, number=1000))
    # print(timeit(f2, number=100))
    # print(timeit(f3, number=1000))
    # print(f2().normalized.as_list())
    # print(a.values)
    # print(a.lower)
    # print(a.categories)
    # a.values = {'test A':'b'}



# -*- coding: utf-8 -*-
'''
iterableSubject class
=============

This `iterable subject` class takes any `subject` object and returns an iterator. 


'''

import pathlib
from typing import Any, Callable, List, Optional, Sequence, Tuple, Union

import reil.subjects as rlsubjects
from reil import stateful
from reil.datatypes import reildata


class IterableSubject(stateful.Stateful):
    '''
    Makes any subject an iterable.

    Attributes
    ----------
        subject:
        save_instances: whether to save instances of the `subject` class or not (Default=False).
        use_existing_instances: whether try to load instances before attempting to create them (Default=True).
        save_path: the path where instances should be saved/ loaded from (Default='./').
        save_prefix: a prefix in instance filenames (Default='').
        instance_counter_start: the index of the first instance of the subject (Default=0).
        instance_counter_end: an int or a list of the last index of the instances (Default=1).
        auto_rewind: whether to rewind after it hits the last instance (Default=False).

    Methods
    -------
        register: register a new agent and return its ID or return ID of an existing agent.
        take_effect: get an action and change the state accordingly.
        reset: reset the state and is_terminated.
    '''

    def __init__(self,
                 subject: rlsubjects.Subject,
                 save_instances: bool = False,
                 use_existing_instances: bool = True,
                 save_path: Union[pathlib.Path, str] = '',
                 save_prefix: str = '',
                 instance_counter_start: int = 0,
                 instance_counter: int = -1,
                 # -1: infinite
                 instance_counter_end: Union[int, List[int]] = -1,
                 end_index: int = 0,
                 auto_rewind: bool = False,
                 **kwargs: Any) -> None:

        self._subject = subject
        self._agent_list = {}
        self._save_instances = save_instances
        self._use_existing_instances = use_existing_instances
        self._save_path = save_path
        self._save_prefix = save_prefix
        self._instance_counter_start = instance_counter_start
        self._instance_counter = instance_counter
        self._end_index = end_index
        self._auto_rewind = auto_rewind

        super().__init__(**kwargs)

        if isinstance(instance_counter_end, int):
            self._instance_counter_end = [instance_counter_end]
        else:
            self._instance_counter_end = instance_counter_end

        if self._instance_counter_end[0] == -1:
            self._stop_check: Callable[[int, int], bool] = lambda current, end: False
        else:
            self._stop_check: Callable[[int, int], bool] = lambda current, end: current > end

    def __iter__(self):
        return self

    def __next__(self) -> Tuple[int, rlsubjects.Subject]:
        self._instance_counter += 1

        try:
            end = self._instance_counter_end[self._end_index]
        except IndexError:
            if self._auto_rewind:
                self._end_index = 0
                self._instance_counter = 1
                end = self._instance_counter_end[self._end_index]
            else:
                raise StopIteration

        if self._stop_check(self._instance_counter, end):
            self._end_index += 1
            self._instance_counter -= 1
            raise StopIteration
        else:
            current_instance = ''.join(
                (self._save_prefix, f'{self._instance_counter:06}'))
            new_instance = True
            if self._use_existing_instances:
                try:
                    self._subject.load(path=self._save_path,
                                       filename=current_instance)
                    new_instance = False
                except FileNotFoundError:
                    self._subject.reset()
            else:
                self._subject.reset()

            if self._save_instances and new_instance:
                self._subject.save(path=self._save_path,
                                   filename=current_instance)

        return (self._instance_counter, self._subject)

    @property
    def state(self) -> reildata.ReilData:
        return self._subject.state

    @property
    def is_terminated(self) -> bool:
        return self._subject.is_terminated

    @property
    def possible_actions(self) -> Sequence[reildata.ReilData]:
        return self._subject.possible_actions

    def take_effect(self, action: reildata.ReilData, _id: Optional[int] = None):
        return self._subject.take_effect(action, _id)

    def reset(self) -> None:
        '''
        Reset the subject.
        This function allows non-iterable use of the defined subject. (e.g. in `env.trajectory()`).
        If you want to reset the iterator itself, use `rewind()` method instead.
        '''
        self._subject.reset()

    def rewind(self) -> None:
        '''
        Rewind the iterator object.
        '''
        self._instance_counter = self._instance_counter_start - 1

    def register(self, agent_name: str) -> int:
        '''
        Registers an agent and returns its ID. If the agent is new, a new ID is generated and the agent_name is added to agent_list.
        \nArguments:
        \n    agent_name: the name of the agent to be registered.
        '''
        try:
            return self._agent_list[agent_name]
        except KeyError:
            try:
                _id = max(self._agent_list.values()) + 1
            except ValueError:
                _id = 1

            self._agent_list[agent_name] = _id
            return _id

    def deregister(self, agent_name: str) -> None:
        '''
        Deegisters an agent given its name.
        \nArguments:
        \n    agent_name: the name of the agent to be registered.
        '''
        self._agent_list.pop(agent_name)

    def __repr__(self) -> str:
        try:
            return f'{self.__class__.__qualname__} -> {self._subject.__repr__()}'
        except AttributeError:
            return self.__class__.__qualname__



# -*- coding: utf-8 -*-
'''
Agent class
===========

This `agent` class is the base class of all agent classes that can learn from
`history`.
'''

import pathlib
from collections import defaultdict
from typing import Any, Dict, Generator, Optional, Tuple, Union, cast

from reil import agents, stateful
from reil.datatypes.reildata import ReilData
from reil.learners.learner import Learner
from reil.utils.exploration_strategies import ExplorationStrategy
from typing_extensions import Literal


class Agent(agents.NoLearnAgent):
    '''
    The base class of all agent classes that learn from history.
    '''

    def __init__(self,
                 learner: Learner,
                 exploration_strategy: ExplorationStrategy,
                 discount_factor: float = 1.0,
                 default_actions: Tuple[ReilData, ...] = (),
                 tie_breaker: Literal['first', 'last', 'random'] = 'random',
                 training_trigger: Literal[
                     'none', 'termination',
                     'state', 'action', 'reward'] = 'termination',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        learner:
            the `Learner` object that does the learning.

        exploration_strategy:
            an `ExplorationStrategy` object that determines
            whether the `action` should be exploratory or not for a given
            `state` at a given `epoch`.

        discount_factor:
            by what factor should future rewards be discounted?

        default_actions:
            a tuple of default actions.

        tie_breaker:
            how to choose the `action` if more than one is candidate
            to be chosen.

        training_trigger:
            When to learn from observations. This arguments is used in
            `observe` method to determine when `learn` method should be called.
            `none` avoids any call to `learn`; `state`, `action` and `reward`
            trigger the `learn` method after receiving their corresponding
            value; `termination` waits until `.close()` method of the generator
            is called.
        '''
        self._tie_breaker: Literal['first', 'last', 'random']

        super().__init__(default_actions, tie_breaker, **kwargs)

        self._learner = learner
        if not 0.0 <= discount_factor <= 1.0:
            self._logger.warning(
                f'{self.__class__.__qualname__} discount_factor should be in'
                f' [0.0, 1.0]. Got {discount_factor}. Set to 1.0.')
        self._discount_factor = min(discount_factor, 1.0)
        self._exploration_strategy = exploration_strategy
        self._training_trigger = training_trigger
        self._history: Dict[int, stateful.History] = defaultdict(list)

    def act(self,
            state: ReilData,
            subject_id: int,
            actions: Optional[Tuple[ReilData, ...]] = None,
            epoch: int = 0) -> ReilData:
        '''
        Return an action based on the given state.

        Arguments
        ---------
        state:
            the state for which the action should be returned.

        subject_id:
            the ID of the `subject` on which action should occur.

        actions:
            the set of possible actions to choose from.

        epoch:
            the epoch in which the agent is acting.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.

        Returns
        -------
        :
            the action
        '''
        if subject_id not in self._entity_list:
            raise ValueError(f'Subject with ID={subject_id} not found.')

        if (self._training_trigger != 'none' and
                self._exploration_strategy.explore(epoch)):
            possible_actions = actions or self._default_actions
            action = self._break_tie(
                possible_actions, self._tie_breaker)
        else:
            action = super().act(state=state, subject_id=subject_id,
                                 actions=actions, epoch=epoch)

        return action

    def reset(self):
        '''Reset the agent at the end of a learning epoch.'''
        super().reset()
        if self._training_trigger != 'none':
            self._learner.reset()

    def load(self, filename: str,
             path: Optional[Union[str, pathlib.Path]] = None) -> None:
        '''
        Load an object from a file.

        Arguments
        ---------
        filename:
            the name of the file to be loaded.

        path:
            the path in which the file is saved.

        Raises
        ------
            ValueError
                Filename is not specified.
        '''
        super().load(filename, path)

        # when loading, self._learner is the object type, not an instance.
        self._learner = self._learner.from_pickle(filename, path)

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.Path]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.Path, str]:
        '''
        Save the object to a file.

        Arguments
        ---------
        filename:
            the name of the file to be saved.

        path:
            the path in which the file should be saved.

        data_to_save:
            a list of variables that should be pickled. If omitted,
            the `agent` is saved completely.

        Returns
        -------
        :
            a `Path` object to the location of the saved file and its name as
            `str`
        '''
        data = list(data_to_save or self.__dict__)
        save_learner = '_learner' in data
        if save_learner:
            data.remove('_learner')

        _path, _filename = super().save(
            filename, path, data_to_save=cast(Tuple, data))

        if save_learner:
            self._learner.save(_filename, _path / 'learner')

        return _path, _filename

    def _prepare_training(self,
                          history: stateful.History) -> agents.TrainingData:
        '''
        Use `history` to create the training set in the form of `X` and `y`
        vectors.

        Arguments
        ---------
        history:
            a `History` object from which the `agent` learns.

        Returns
        -------
        :
            a `TrainingData` object that contains `X` and 'y` vectors


        :meta public:
        '''
        raise NotImplementedError

    def learn(self, history: stateful.History) -> None:
        '''
        Learn using history.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` whose history is being used for learning.

        next_state:
            The new `state` of the `subject` after taking `agent`'s action.
            Some methods
        '''
        if history is not None:
            X, Y = self._prepare_training(history)
        else:
            X, Y = cast(agents.TrainingData, ([], []))

        if X:
            self._learner.learn(X, Y)

    def observe(self, subject_id: int, stat_name: Optional[str],  # noqa: C901
                ) -> Generator[Union[ReilData, None], Any, None]:
        '''
        Create a generator to interact with the subject (`subject_id`).
        Extends `NoLearnAgent.observe`.

        This method creates a generator for `subject_id` that
        receives `state`, yields `action` and receives `reward`
        until it is closed. When `.close()` is called on the generator,
        `statistics` are calculated.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` on which action happened.

        stat_name:
            The name of the `statistic` that should be computed at the end of
            each trajectory.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.
        '''
        if subject_id not in self._entity_list:
            raise ValueError(f'Subject with ID={subject_id} not found.')

        history: stateful.History = []
        new_observation = stateful.Observation()
        while True:
            try:
                new_observation = stateful.Observation()
                temp = yield
                new_observation.state = cast(ReilData, temp['state'])
                actions: Tuple[ReilData, ...] = temp['actions']
                epoch: int = temp['epoch']

                if self._training_trigger == 'state':
                    self.learn([history[-1], new_observation])

                if actions is not None:
                    new_observation.action = self.act(
                        state=new_observation.state, subject_id=subject_id,
                        actions=actions, epoch=epoch)

                    if self._training_trigger == 'action':
                        self.learn([history[-1], new_observation])

                    new_observation.reward = (yield new_observation.action)

                    history.append(new_observation)

                    if self._training_trigger == 'reward':
                        self.learn(history[-2:])
                else:
                    yield

            except GeneratorExit:
                if new_observation.reward is None:  # terminated early!
                    history.append(new_observation)

                if self._training_trigger == 'termination':
                    self.learn(history)

                if stat_name is not None:
                    self.statistic.append(stat_name, subject_id)

                return



# -*- coding: utf-8 -*-
'''
LegacyReilData class
==================

A data type used for state and action variables.


'''

# from __future__ import annotations
from numbers import Number
from typing import (Any, Callable, Dict, Iterator, List, Optional, Sequence,
                    Union)


class LegacyReilData(dict):
    def __new__(cls, value=[0], **kwargs):
        obj = super().__new__(cls)

        obj._lazy = True
        obj._value = {}
        obj._is_numerical = {}
        obj._normalizer = {}
        obj._categories = {}
        obj._lower = {}
        obj._upper = {}
        obj._normal_form = {}

        return obj

    def __init__(self, value: Optional[Union[Sequence, dict]] = [0],
                 lower: Optional[Union[Sequence, dict]] = None,
                 upper: Optional[Union[Sequence, dict]] = None,
                 categories: Optional[Union[Sequence, dict]] = None,
                 is_numerical: Optional[Union[Sequence[bool], Dict[Any, bool]]] = None,
                 normalizer: Optional[Union[Callable[[dict], Number], Callable[[dict], List[Number]],
                                            Dict[Any, Callable[[dict], Number]], Dict[Any, Callable[[dict], List[Number]]]]] = None,
                 lazy_evaluation: Optional[bool] = False) -> None:
        '''
        Create an LegacyReilData instance.

        Attributes:
            value: value to store as a list or a dictionary
            lower: minimum values for numerical components
            upper: maximum values for numerical components
            categories: set of categories for categorical components
            is_numerical: whether each component is numerical (True) or categorical (False)
            normalizer: a function that normalizes the respective component
            lazy_evaluation: whether to store normalized values or compute on-demand (Default: False)
        '''
        self._value = {}

        self.value = value

        if is_numerical is not None:
            self.is_numerical = is_numerical

        if lower is not None:
            self.lower = lower

        if upper is not None:
            self.upper = upper

        if categories is not None:
            self.categories = categories

        if normalizer is not None:
            self.normalizer = normalizer

        self._normalizer_lambda_func = lambda x: (x['value']-x['lower'])/(
            x['upper']-x['lower']) if x['is_numerical'] else list(int(x_i == x['value']) for x_i in x['categories'])

        if lazy_evaluation is not None:
            self._lazy = lazy_evaluation

        if not self._lazy:
            self._normalized = self._normalize()

    @property
    def value(self) -> "LegacyReilData":
        return self._value

    @value.setter
    def value(self, v: Union[Sequence, dict]) -> None:
        '''
        Sets the value of the LegacyReilData instance.

        Attributes:
            v: value to store as a list or a dictionary. If a list is provided,
            it is stored as `value`. To have named values, use dictionary.
        '''
        try:
            for i, v_i in v.items():
                self.__setitem__(i, v_i)
        except AttributeError:
            self.__setitem__(None, v)  # .copy())

    @property
    def lower(self) -> Union[list, dict]:
        '''
        Return the lower bound
        '''
        return self._lower

    @lower.setter
    def lower(self, value: Union[Sequence, dict]) -> None:
        '''
        Set the lower bound

        Raises ValueError if the provided lower bound is greater than the currently stored value. 
        '''
        def check_min(v, new_min):
            try:
                minimum = min(v)
            except TypeError:
                minimum = v

            if minimum < new_min:
                raise ValueError(
                    f'The provided lower bound ({new_min}) is greater than current smallest number ({minimum}).\n{self.__repr__()}')

            return new_min

        try:
            for i, val in value.items():
                if self._is_numerical[i]:
                    self._lower[i] = check_min(self._value[i], val)

        except AttributeError:
            if self._is_numerical:
                self._lower = check_min(self._value, value)

        if not self._lazy:
            self._normalized = self._normalize()

    @property
    def upper(self) -> Union[list, dict]:
        '''
        Return the upper bound
        '''
        return self._upper

    @upper.setter
    def upper(self, value: Union[Sequence, dict]) -> None:
        '''
        Set the upper bound

        Raises ValueError if the provided upper bound is less than the currently stored value. 
        '''
        def check_max(v, new_max):
            try:
                maximum = max(v)
            except TypeError:
                maximum = v

            if maximum > new_max:
                raise ValueError(
                    f'The provided upper bound ({new_max}) is less than current greatest number ({maximum}).\n{self.__repr__()}')

            return new_max

        try:
            for i, val in value.items():
                if self._is_numerical[i]:
                    self._upper[i] = check_max(self._value[i], val)

        except AttributeError:
            if self._is_numerical:
                self._upper = check_max(self._value, value)

        if not self._lazy:
            self._normalized = self._normalize()

    @property
    def categories(self) -> Union[list, dict]:
        '''
        Return the list of all categories.
        '''
        return self._categories

    @categories.setter
    def categories(self, value: Union[Sequence, dict]) -> None:
        '''
        Set the categories (for categorical components only)
        '''
        try:
            for i, val in value.items():
                if not self._is_numerical[i]:
                    self._categories[i] = val
        except AttributeError:
            if not self._is_numerical:
                self._categories = value

        if not self._lazy:
            self._normalized = self._normalize()

    @property
    def is_numerical(self) -> Union[list, dict]:
        '''
        Return a boolean dataframe that shows if each component is numerical or not.
        '''
        return self._is_numerical

    @is_numerical.setter
    def is_numerical(self, value: Union[Sequence, dict]) -> None:
        # If it was numerical and the user assigns numerical (True and True) then it remains numerical.
        # But if it is not numerical, then we cannot change it to numerical.
        try:
            for i, val in value.items():
                self._is_numerical[i] = self._is_numerical[i] and val
        except AttributeError:
            self._is_numerical = self._is_numerical and value

        if not self._lazy:
            self._normalized = self._normalize()

    @property
    def normalizer(self) -> Union[list, dict]:
        return self._normalizer

    @normalizer.setter
    def normalizer(self, value: Union[Sequence, dict]) -> None:
        try:
            for i, val in value.items():
                self._normalizer[i] = val
        except AttributeError:
            self._normalizer = value

        if not self._lazy:
            self._normalized = self._normalize()

    def as_list(self) -> list:
        ''' return the value as list.'''
        try:
            return self.__remove_nestings(self._value.values())
        except AttributeError:
            return self._value

    def as_reildata_array(self) -> List["LegacyReilData"]:
        ''' return the value as a list of LegacyReilData.'''
        try:
            array = [LegacyReilData(value=self._value[key],
                            lower=self._lower[key],
                            upper=self._upper[key],
                            categories=self._categories[key],
                            is_numerical=self._is_numerical[key],
                            lazy_evaluation=self._lazy) for key in self._value.keys()]
        except AttributeError:
            array = [LegacyReilData(value=v,
                            lower=self._lower,
                            upper=self._upper,
                            categories=self._categories,
                            is_numerical=self._is_numerical,
                            lazy_evaluation=self._lazy) for v in self._value]

        return array

    def __remove_nestings(self, l: list) -> list:
        output = []
        for i in l:
            if isinstance(i, (list, dict)):
                output += self.__remove_nestings(i)
            else:
                output.append(i)

        return output

    def _normalize(self, keys: Optional[Sequence] = None) -> list:
        temp = []
        try:
            keys_list = keys if keys is not None else self._value.keys()
            for i in keys_list:
                if self._normal_form[i] is None:
                    if self._normalizer[i] is None:
                        func = self._normalizer_lambda_func
                    else:
                        func = self._normalizer[i]

                    try:
                        self._normal_form[i] = func({'value': self._value[i],
                                                     'lower': self._lower[i],
                                                     'upper': self._upper[i],
                                                     'categories': self._categories[i],
                                                     'is_numerical': self._is_numerical[i]})
                    except TypeError:
                        self._normal_form[i] = list(func({'value': x,
                                                          'lower': self._lower[i],
                                                          'upper': self._upper[i],
                                                          'categories': self._categories[i],
                                                          'is_numerical': self._is_numerical[i]}) for x in self._value[i])
                    except ZeroDivisionError:
                        self._normal_form[i] = [1]
                temp.append(self._normal_form[i])
        except AttributeError:
            if self._normal_form is None:
                if self._normalizer is None:
                    func = self._normalizer_lambda_func
                else:
                    func = self._normalizer

                try:
                    try:
                        self._normal_form = list(func({'value': x,
                                                       'lower': self._lower,
                                                       'upper': self._upper,
                                                       'categories': self._categories,
                                                       'is_numerical': self._is_numerical}) for x in self._value)
                    except TypeError:
                        self._normal_form = [func({'value': self._value,
                                                   'lower': self._lower,
                                                   'upper': self._upper,
                                                   'categories': self._categories,
                                                   'is_numerical': self._is_numerical})]
                except ZeroDivisionError:
                    self._normal_form = [1]

            temp = self._normal_form

        return self.__remove_nestings(temp)

    def normalize(self) -> "LegacyReilData":
        '''
        Normalize values.

        This function uses max and min for numericals and categories for categoricals to turn them into [0, 1] values.
        '''
        if self._lazy:
            return LegacyReilData(self._normalize(), lower=0, upper=1, lazy_evaluation=True)
        else:
            return LegacyReilData(self._normalized, lower=0, upper=1, lazy_evaluation=True)

    def __setitem__(self, key: Any, value: Any) -> "LegacyReilData":
        if key is None:  # complete list
            if isinstance(self._value, dict):  # Go from dict to list
                self._is_numerical = None
                self._lower = None
                self._upper = None
                self._categories = None
                self._normalizer = None

            old_lower = self._lower
            old_upper = self._upper
            old_categories = self._categories
            old_normalizer = self._normalizer

            if not hasattr(value, '__iter__') or isinstance(value, str):
                temp = [value]
            else:
                temp = value

            if self._categories is not None:
                for t in temp:
                    if t not in self._categories:
                        self._lower = old_lower
                        self._upper = old_upper
                        self._categories = old_categories
                        self._normalizer = old_normalizer
                        raise ValueError(
                            f'{t} is not found in {self._categories}.')
            elif self._lower is not None:
                for t in temp:
                    if not (self._lower <= t <= self._upper):
                        self._lower = old_lower
                        self._upper = old_upper
                        self._categories = old_categories
                        self._normalizer = old_normalizer
                        raise ValueError(
                            f'{t} is outside the range [{self._lower}, {self._upper}].')
            else:
                self._is_numerical = all(isinstance(v_i, Number)
                                         for v_i in temp)
                if self._is_numerical:
                    self._lower = min(temp)
                    self._upper = max(temp)
                    self._categories = None
                else:
                    self._lower = None
                    self._upper = None
                    self._categories = temp

            self._value = temp
            self._normal_form = None

        elif isinstance(key, slice):  # slice of a list
            if isinstance(self._value, dict):
                raise TypeError(
                    'Cannot use slice for a LegacyReilData instance of type dict.')

            if self._categories is not None:
                for t in value:
                    if t not in self._categories:
                        raise ValueError(
                            f'{t} is not found in {self._categories}.')
            elif self._lower is not None:
                for t in value:
                    if not (self._lower <= t <= self._upper):
                        raise ValueError(
                            f'{t} is outside the range [{self._lower}, {self._upper}].')
            else:
                raise RuntimeError(
                    'LegacyReilData is corrupted! No categories, lower or upper attributes found!')

            self._value[key] = value
            self._normal_form = None
        elif isinstance(self._value, dict):
            if not hasattr(value, '__iter__') or isinstance(value, str):
                temp = [value]
            else:
                temp = value

            if key not in self._value.keys():
                self._normalizer[key] = None
                self._is_numerical[key] = all(
                    isinstance(v_i, Number) for v_i in temp)
                if self._is_numerical[key]:
                    self._lower[key] = min(temp)
                    self._upper[key] = max(temp)
                    self._categories[key] = None
                else:
                    self._lower[key] = None
                    self._upper[key] = None
                    self._categories[key] = temp

            if self._is_numerical[key]:
                for v in temp:
                    if not (self._lower[key] <= v <= self._upper[key]):
                        raise ValueError(
                            f'{v} is outside the range [{self._lower[key]}, {self._upper[key]}].')
            else:
                if value not in self._categories[key]:
                    for v in temp:
                        if v not in self._categories[key]:
                            raise ValueError(
                                f'{value} is not found in {self._categories[key]}.')

            self._value[key] = value
            self._normal_form[key] = None

        else:  # index of a list
            if self._categories is not None:
                if value not in self._categories:
                    raise ValueError(
                        f'{value} is not found in {self._categories}.')
            elif self._lower is not None:
                if not (self._lower <= value <= self._upper):
                    raise ValueError(
                        f'{value} is outside the range [{self._lower}, {self._upper}].')
            else:
                raise RuntimeError(
                    'LegacyReilData is corrupted! No categories, lower or upper attributes found!')

            self._value[key] = value
            self._normal_form = None

        if not self._lazy:
            self._normalized = self._normalize()

        return value

    def __getitem__(self, key: Any) -> "LegacyReilData":
        try:
            return LegacyReilData(self._value[key],
                          lower=self._lower[key],
                          upper=self._upper[key],
                          categories=self._categories[key],
                          is_numerical=self._is_numerical[key],
                          normalizer=self._normalizer[key],
                          lazy_evaluation=self._lazy)
        except (TypeError, IndexError):
            if isinstance(key, Number):
                return self._value[key]
            else:
                return LegacyReilData(self._value[key],
                              lower=self._lower,
                              upper=self._upper,
                              categories=self._categories,
                              is_numerical=self._is_numerical,
                              normalizer=self._normalizer,
                              lazy_evaluation=self._lazy)

    def __delitem__(self, key: Any) -> None:
        del self._value[key]

        try:
            del self._lower[key]
            del self._upper[key]
        except KeyError:
            del self._categories[key]

        try:
            del self._is_numerical[key]
            del self._normalizer[key]
            del self._normalized[key]
        except KeyError:
            pass

    def clear(self) -> Union[list, dict]:
        return self._value.clear()

    def has_key(self, k: Any) -> bool:
        return k in self._value.keys()

    def update(self, kwargs: Union["LegacyReilData", list, dict]) -> "LegacyReilData":
        try:
            if isinstance(kwargs._value, list):
                if self.is_numerical:
                    if min(kwargs._value) >= self.lower and max(kwargs._value) <= self.upper:
                        self._value += kwargs._value
                else:
                    if all(item in self.categories for item in kwargs._value):
                        self._value += kwargs._value

                return self.value

        except AttributeError:
            pass

        for k, v in kwargs.items():
            self.__setitem__(k, v)

        return self.value

    def keys(self) -> Any:
        try:
            return self._value.keys()
        except AttributeError:
            try:
                return slice(0, len(self._value))
            except TypeError:
                return 0

    def values(self) -> "LegacyReilData":
        try:
            return self._value.values()
        except AttributeError:
            return self._value

    def items(self) -> enumerate:
        try:
            return self._value.items()
        except AttributeError:
            try:
                return enumerate(self._value)
            except TypeError:
                return enumerate([self._value])

    def pop(self, *args: Any) -> "LegacyReilData":
        return self._value.pop(*args)

    def __contains__(self, item: Any) -> bool:
        return item in self._value

    def __iter__(self) -> Iterator:
        return iter(self._value)

    def __add__(self, other: Union["LegacyReilData", list, dict]) -> "LegacyReilData":
        temp = LegacyReilData(value=self._value,
                      lower=self.lower,
                      upper=self.upper,
                      categories=self.categories,
                      is_numerical=self.is_numerical,
                      normalizer=self._normalizer
                      )
        temp.update(other)
        return temp

    def __iadd__(self, other: Union["LegacyReilData", list, dict]) -> "LegacyReilData":
        self.update(other)
        return self

    def __eq__(self, other: "LegacyReilData") -> bool:
        try:
            return (self.value == other.value).bool() and (
                ((self.upper == other.upper).bool() and (self.lower == other.lower).bool()) if self.is_numerical.bool() else
                (self.categories == other.categories).bool())
        except AttributeError:
            return (self.value == other.value) and (
                ((self.upper == other.upper) and (self.lower == other.lower)) if self.is_numerical else
                (self.categories == other.categories))

    def __ge__(self, other: Union["LegacyReilData", list, dict]) -> bool:
        if isinstance(other, LegacyReilData):
            other_value = other.value
        else:
            other_value = other

        try:
            return (self.value >= other_value).bool()
        except AttributeError:
            return (self.value >= other_value)

    def __gt__(self, other: Union["LegacyReilData", list, dict]) -> bool:
        if isinstance(other, LegacyReilData):
            other_value = other.value
        else:
            other_value = other

        try:
            return (self.value > other_value).bool()
        except AttributeError:
            return (self.value > other_value)

    def __le__(self, other: Union["LegacyReilData", list, dict]) -> bool:
        if isinstance(other, LegacyReilData):
            other_value = other.value
        else:
            other_value = other

        try:
            return (self.value <= other_value).bool()
        except AttributeError:
            return (self.value <= other_value)

    def __lt__(self, other: Union["LegacyReilData", list, dict]) -> bool:
        if isinstance(other, LegacyReilData):
            other_value = other.value
        else:
            other_value = other

        try:
            return (self.value < other_value).bool()
        except AttributeError:
            return (self.value < other_value)

    def __ne__(self, other: Union["LegacyReilData", list, dict]) -> bool:
        if isinstance(other, LegacyReilData):
            other_value = other.value
        else:
            other_value = other

        try:
            return (self.value != other_value).bool()
        except AttributeError:
            return (self.value != other_value)

    def __format__(self, formatstr: str) -> str:
        try:
            return '[' + ', '.join(format(i, formatstr) for i in self.value) + ']'
        except TypeError:
            return format(self.value, formatstr)
        except AttributeError:
            return False

    def __len__(self) -> int:
        try:
            return len(self._value)
        except TypeError:
            return 1

    def __hash__(self) -> int:
        return self._value.__hash__()

    def __repr__(self) -> str:
        return f'[{str(self.value)}]\nlower={str(self._lower)}\nupper={str(self._upper)}\ncategories={str(self._categories)}'

    def __str__(self) -> str:
        return str(self._value)


if __name__ == '__main__':
    # d = LegacyReilData([1, 2, 3], lower=1, upper=10)
    # print(d._value)
    # print(d._normalized)
    # d.value = 10
    # print(d._normalized)

    # d = LegacyReilData({'a': [10, 20], 'b': [30, 10, 5, 40], 'c': 50, 'd': 'hello'},
    #            lower={'a': 1, 'b': 2, 'c': 3, 'd': 'a'})
    # print(d._value)
    # print(d._normalized)
    # d.upper = {'b': 100, 'c': 50, 'a': 20, 'd': 'zzzzzz'}
    # d.lower = {'b': -10, 'c': 0, 'a': 1, 'd': '0'}
    # print(d._value)
    # print(d._normalized)
    # # d.is_numerical={'a': False}
    # for temp in d.as_reildata_array():
    #     print(temp)

    # print(d.normalize())
    d = LegacyReilData([1, 2, 3], lazy_evaluation=True)
    print(d.value)
    print(d.normalize())
    print(d.normalize())
    d += [1, 2, 3]
    print(d.normalize())
    print(d.as_reildata_array())
    print(d == d)
    d += LegacyReilData([1.5, 2.5, 3], lower=0)
    print(d)

    d = LegacyReilData({'a': 1, 'b': 2, 'c': 3},
               lower={'a': 0, 'b': 0},
               upper={'a': 10, 'b': 10},
               categories={'c': (1, 2, 3)},
               is_numerical={'c': False})

    print(d+LegacyReilData({'a': 5, 'c': 1}, is_numerical={'a': True,
                                                   'c': False}, lazy_evaluation=True))

    d1 = LegacyReilData(['a', 'b', 'c'], categories=['a', 'b', 'c'])
    assert d1.value == ['a', 'b', 'c']
    print(d1.value)
    print(d1.normalize())
    print(d1.as_reildata_array())
    d = LegacyReilData({'tuples': [(1, 1), (1, 2), (1, 3)], 'ints': 1})
    print(d.value)
    print(d.normalize())
    d_temp = d['tuples']
    print(d_temp[0])
    print(d.as_reildata_array()[0].normalize().as_list())



# -*- coding: utf-8 -*-
'''
healthcare subjects module for reinforcement learning
=====================================================

This module provides different healthcare subjects in reinforcement learning
context.

Classes
-------
Patient:
    The base class of all healthcare subject classes

WarfarinPatientRavvaz:
    A warfarin patient model with features and parameters of
    Ravvaz et al. 2016.

CancerModel:
    A 4-ordinary differential equation model of cancer.
    (uses legacy ValueSet instead of ReilData)

ConstrainedCancerModel:
    A constrained version of CancerModel.
    (uses legacy ValueSet instead of ReilData)

Warfarin:
    A `Subject` for warfarin that uses `WarfarinPatientRavvaz` and
    `healthcare.hamberg_pkpd`.
'''

# TODO: update CancerModel
# TODO: update ConstrainedCancerModel

from .patient import Patient  # noqa: W0611
from .warfarin_patient_ravvaz import WarfarinPatientRavvaz  # noqa: W0611
# from .cancer_model import CancerModel  # noqa: W0611
# from .constrained_cancer_model import ConstrainedCancerModel  # noqa: W0611
from .warfarin import Warfarin  # noqa: W0611



# -*- coding: utf-8 -*-
'''
datatypes module for reinforcement learning
===========================================

This module contains datatypes used in `reil`

Submodules
----------
buffers:
    A module that contains different types of `Buffers` used in `reil`
    package.

Classes
-------
Feature:
    A datatype that accepts initial value and feature generator and
    generates new values.

Entity:
    A datatype to specify `agent` or `subject` information. Used in
    `InteractionProtocol`.

InteractionProtocol:
    A datatype to specifies how an `agent` and a `subject`
    interact in an `environment`.

ReilData:
    The main datatype used to communicate `state`s, `action`s, and `reward`s,
    between objects in `reil`. `ReilData` is basically a tuple that contains
    instances of `BaseData`, `CategoricalData`, and `NumericalData`.

BaseData:
    An immutable dataclass that accepts name, value, and normalizer.

CategoricalData:
    Extends `BaseData` by adding a list of categories.

NumericalData:
    Extends `BaseData` by adding lower and upper bounds for the value.

PrimaryComponent:
    A datatype that is being used mostly by children of `Stateful` to include
    a `PrimaryComponent`, e.g. a state. It allows defining different
    definitions for the component, and call the instance to calculate them.

SecondayComponent:
    A datatype that is being used mostly by children of `Subject` to include
    a `SecondayComponent`, e.g. a statistic or a reward. It allows defining
    different definitions for the component, and call the instance to calculate
    them.
'''

from .reildata import (BaseData, CategoricalData, NumericalData,  # noqa: W0611
                       ReilData)

from .feature import Feature, FeatureType  # noqa: W0611
from .interaction_protocol import Entity, InteractionProtocol  # noqa: W0611

from .components import (PrimaryComponent, SecondayComponent,  # noqa: W0611
                         Statistic, MockStatistic, SubComponentInfo)

import reil.datatypes.buffers  # noqa: W0611



# -*- coding: utf-8 -*-
'''
cancer_model class
==================

This `cancer_model` class implements a four-state nonlinear cancer chemotherapy model. 


'''

from ..legacy import ValueSet
from .cancer_model import CancerModel


class ConstrainedCancerModel(CancerModel):
    '''
    Four-state nonlinear cancer chemotherapy model with constraint on drug dose.
    
    Attributes
    ----------
        state: the state of the subject as a ValueSet.
        is_terminated: whether the subject is finished or not.
        possible_actions: a list of possible actions.

    Methods
    -------
        register: register a new agent and return its ID or return ID of an existing agent.
        take_effect: get an action and change the state accordingly.
        reset: reset the state and is_terminated.
    '''
    def __init__(self, **kwargs):
        self._drug_cap = lambda x: kwargs.get('u_max', 0)
        if 'drug_cap' in kwargs:
            self.set_params(drug_cap=kwargs['drug_cap'])
        super().__init__(**kwargs)

        # The following code is just to suppress debugger's undefined variable errors!
        # These can safely be deleted, since all the attributes are defined using set_params!
        if False:
            self._drug_cap=lambda x, day: self._u_max

        self._x['drug_cap'] = self._drug_cap(self._x)

    @property
    def possible_actions(self):
        return ValueSet([self._x['drug_cap']*x/self._u_steps for x in range(0, self._u_steps+1)], min=0, max=self._u_max, 
                        binary=lambda x: (int(x * self._u_steps // self._u_max), self._u_steps+1)).as_valueset_array()

    def take_effect(self, action, _id=None):
        r = CancerModel.take_effect(self, action, _id)
        self._x['drug_cap'] = self._drug_cap(self._x)
        return r

    def reset(self):
        CancerModel.reset(self)
        self._x['drug_cap'] = self._drug_cap(self._x)

    def __repr__(self):
        try:
            return f"ConstrainedCancerModel: [day: {self._x['day']}, N: {self._x['normal_cells']}, T: {self._x['tumor_cells']}, " \
                   f"N: {self._x['immune_cells']}, C: {self._x['drug']}, drug cap: {self._x.get('drug_cap','N/A')}]"
        except:
            return 'ConstrainedCancerModel'



# -*- coding: utf-8 -*-
'''
HambergPKPD class
=================

A PK/PD model proposed by Hamberg et al. (2007).
DOI: 10.1038/sj.clpt.6100084
'''
import math
from collections import namedtuple
from reil.datatypes.feature import Feature
from typing import Any, Callable, Dict, List, Union, cast

import numpy as np
from reil.subjects.healthcare.mathematical_models import HealthMathModel

DoseEffect = namedtuple('DoseEffect', ['dose', 'Cs'])


class HambergPKPD(HealthMathModel):
    '''
    Hamberg PK/PD model for warfarin.
    '''
    def __init__(self, randomized: bool = True, cache_size: int = 30) -> None:
        """
        Arguments
        ---------
        randomized:
            Whether to have random effects in patient response to warfarin.

        cache_size:
            Size of the cache used to store pre-computed values needed for
            INR computation.
        """
        self._randomized = randomized
        self._d2h = 24

        self._gamma = 0.424  # no units
        self._Q = 0.131    # (L/h)
        self._lambda = 3.61
        self._E_max = 1	 # no units

        self._cache_size = cache_size
        self._cached_cs = []

    def setup(self, **arguments: Union[Feature[str], Feature[float]]) -> None:
        '''
        Set up the model.

        Arguments
        ---------
        arguments:
            `Feature` instances required to setup the model.

        Notes
        -----
        This model requires `age`, `CYP2C9`, `MTT_1`, `MTT_2`, `EC_50`,
        `cyp_1_1`, `V1`, and `V2`. The genotype of `VKORC1` is not directly
        used in this implementation. Instead, one should use it to generate
        `EC_50`.

        Raises
        ------
        ValueError:
            `CYP2C9` is not one of the acceptable values:
            *1/*1, *1/*2, *1/*3, *2/*2, *2/*3, *3/*3
        '''
        age = cast(float, arguments['age'].value)
        CYP2C9 = cast(str, arguments['CYP2C9'].value)

        # VKORC1 is not used inside this implementation of Hamberg. It
        # specifies the distribution of EC_50 which is being determined
        # in WarfarinPatient class.

        MTT_1 = cast(float, arguments['MTT_1'].value)
        MTT_2 = cast(float, arguments['MTT_2'].value)

        self._EC_50_gamma = cast(
            float, arguments['EC_50'].value) ** self._gamma

        cyp_1_1 = cast(float, arguments['cyp_1_1'].value)
        V1 = cast(float, arguments['V1'].value)
        V2 = cast(float, arguments['V2'].value)

        ktr1 = 6/MTT_1  # 1/hours; changed from 1/MTT_1
        ktr2 = 1/MTT_2  # 1/hours
        self._ktr = np.array([0.0] + [ktr1] * 6 + [0.0, ktr2])

        Q = 0.131    # (L/h)

        # bioavilability fraction 0-1 (from: "Applied Pharmacokinetics &
        # Pharmacodynamics 4th edition, p.717", some other references)
        F = 0.9

        self._ka = 2  # absorption rate (1/hr)

        temp = {'*1/*1': 0.0,
                '*1/*2': 0.315,
                '*1/*3': 0.453,
                '*2/*2': 0.722,
                '*2/*3': 0.69,
                '*3/*3': 0.852}

        if CYP2C9 not in temp:
            raise ValueError('The CYP2C9 genotype not recognized fool!')

        CL_s = 1.0 - (0.0091 * (age - 71) if age > 71 else 0)
        CL_s *= cyp_1_1 * (1 - temp[CYP2C9])

        k12 = Q / V1
        k21 = Q / V2
        k10 = CL_s / V1

        b = k10 + k21 + k12
        c = k10 * k21
        self._alpha = (b + math.sqrt(b ** 2 - 4*c)) / 2
        self._beta = (b - math.sqrt(b ** 2 - 4*c)) / 2

        kaF_2V1 = (self._ka * F / 2) / V1
        self._coef_alpha = max(
            0.0, ((k21 - self._alpha)
                  / ((self._ka - self._alpha)*(self._beta - self._alpha)))
        ) * kaF_2V1
        self._coef_beta = max(
            0.0, ((k21 - self._beta)
                  / ((self._ka - self._beta)*(self._alpha - self._beta)))
        ) * kaF_2V1
        self._coef_k_a = max(
            0.0, ((k21 - self._ka)
                  / ((self._ka - self._alpha)*(self._ka - self._beta)))
        ) * kaF_2V1

        self._dose_records: Dict[int, DoseEffect] = {}
        self._total_cs = np.array(
            [0.0] * self._cache_size * self._d2h)  # hourly
        self._list_of_INRs = [0.0] * (self._cache_size + 1)  # daily
        self._err_list = []  # daily
        self._err_ss_list = []  # daily
        self._exp_e_INR_list = []  # daily
        self._last_computed_day: int = 0

        temp_cs_generator = self._CS_function_generator(0, 1.0)
        self._cached_cs = [temp_cs_generator(t)
                           for t in range(self._cache_size * self._d2h)]

        self._A = np.array([0.0] + [1.0] * 8)
        self._list_of_INRs[0] = self._INR(self._A, 0)

    def run(self, **inputs: Any) -> Dict[str, Any]:
        '''
        Run the model.

        Arguments
        ---------
        inputs:
            - A dictionary called "dose" with days for each dose as keys and
              the amount of dose as values.
            - A list called "measurement_days" that shows INRs of which days
              should be returned.

        Returns
        -------
        :
            A dictionary with keyword "INR" and a list of doses for the
            specified days.
        '''
        self.dose = inputs.get('dose', {})

        return {'INR': self.INR(inputs.get('measurement_days', []))}

    @property
    def dose(self) -> Dict[int, float]:
        '''
        Return doses for each day.

        Returns
        -------
        :
            A dictionary with days as keys and doses as values.
        '''
        return dict((t, info.dose)
                    for t, info in self._dose_records.items())

    @dose.setter
    def dose(self, dose: Dict[int, float]) -> None:
        '''
        Add warfarin doses at the specified days.

        Arguments
        ---------
        dose:
            A dictionary with days as keys and doses as values.
        '''
        # if a dose is added ealier in the list, INRs should be updated all
        # together because the history of "A" array is not kept.
        try:
            if self._last_computed_day > min(dose.keys()):
                self._last_computed_day = 0
        except ValueError:  # no doses
            pass

        for t, d in dose.items():
            if d != 0.0:
                h = t * self._d2h
                if t in self._dose_records:
                    self._total_cs -= (np.array([0.0]*h + self._cached_cs[:-h])
                                       * self._dose_records[t])

                self._dose_records[t] = DoseEffect(
                    d, self._CS_function_generator(h, d))

                try:
                    self._total_cs += np.array(
                        [0.0]*h +
                        self._cached_cs[:-h]
                    )[:self._cache_size * self._d2h] * d
                except ValueError:  # _t == 0
                    self._total_cs += np.array(self._cached_cs) * d

    def INR(self, measurement_days: Union[int, List[int]]) -> List[float]:
        '''
        Compute INR values for the specified days.

        Arguments
        ---------
        measurement_days:
            One of a list of all days for which INR should be computed.

        Returns
        -------
        :
            A list of INRs for the specified days.
        '''
        if isinstance(measurement_days, int):
            days = [measurement_days]
        else:
            days = measurement_days

        if self._last_computed_day == 0:
            self._A = np.array([0.0] + [1.0] * 8)

        max_days = max(days)
        self._list_of_INRs.extend(
            [0.0] * (max_days - len(self._list_of_INRs) + 1))

        for d in range(self._last_computed_day, max_days):
            for i in range(d * 24, (d + 1) * 24):
                self._A[0] = self._A[7] = self._inflow(i)
                self._A[1:] += self._ktr[1:] * (self._A[0:-1] - self._A[1:])
            self._list_of_INRs[d + 1] = self._INR(self._A, d + 1)

        self._last_computed_day = max_days

        return [self._list_of_INRs[i] for i in days]

    def _CS_function_generator(
            self, t_dose: int, dose: float) -> Callable[[int], float]:
        '''
        Generate a Cs function.

        Arguments
        ---------
        t_dose:
            The day in which the dose is administered.

        dose:
            The value of the dose administered.

        Returns
        -------
        :
            A function that gets the day and return that day's
            warfarin concentration.

        Notes
        -----
        To speed up the process, the generated function uses a pre-computed
        cache of concentrations and only computes the concentration
        if the requested day is beyond the cached range.
        '''
        cached_cs_temp = [dose * cs for cs in self._cached_cs]

        def Cs(t: int) -> float:
            '''
            Get a day and return its warfarin concentration.

            Arguments
            ---------
            t:
                The day for which concentration value is needed.

            Returns
            -------
            :
                Warfarin concentration
            '''
            if t <= t_dose:
                return 0.0

            try:
                return cached_cs_temp[t - t_dose]
            except IndexError:
                return (self._coef_alpha * math.exp(
                    -self._alpha * (t - t_dose)) +
                    self._coef_beta * math.exp(
                    -self._beta * (t - t_dose)) +
                    self._coef_k_a * math.exp(
                    -self._ka * (t - t_dose))) * dose

        return Cs

    def _err(self, t: int, ss: bool = False) -> float:
        '''
        Generate error term for the requested day.

        Arguments
        ---------
        t:
            The day for which the error is requested.

        ss:
            Whether the dosing has reached the steady-state.

        Returns
        -------
        :
            The error value.

        Notes
        -----
        To speed up the process and generate reproducible results in each run,
        the errors are cached in batches.
        For each call of the function, the cached error is returned. If
        the `day` is beyond the cached range, a new range of error values
        are generated and added to the cache.
        '''
        if self._randomized:
            index_0 = t // self._cache_size
            index_1 = t % self._cache_size
            e_list = self._err_ss_list if ss else self._err_list
            try:
                return e_list[index_0][index_1]
            except IndexError:
                missing_rows = index_0 - len(e_list) + 1
                stdev = 0.30 if ss else 0.09
                for _ in range(missing_rows):
                    e_list.append(np.exp(np.random.normal(  # type:ignore
                        0, stdev, self._cache_size)))

            return e_list[index_0][index_1]
        else:
            return 1.0

    def _exp_e_INR(self, t: int) -> float:
        '''
        Generate exp(error) term of INR for the requested day.

        Arguments
        ---------
        t:
            The day for which the error is requested.

        Returns
        -------
        :
            The error value.

        Notes
        -----
        To speed up the process and generate reproducible results in each run,
        the errors are cached in batches.
        For each call of the function, the cached error is returned. If
        the `day` is beyond the cached range, a new range of error values
        are generated and added to the cache.
        '''
        if self._randomized:
            index_0 = t // self._cache_size
            index_1 = t % self._cache_size
            try:
                return self._exp_e_INR_list[index_0][index_1]
            except IndexError:
                missing_rows = index_0 - len(self._exp_e_INR_list) + 1
                for _ in range(missing_rows):
                    self._exp_e_INR_list.append(
                        np.exp(np.random.normal(  # type:ignore
                            0, 0.0325, self._cache_size)))

            return self._exp_e_INR_list[index_0][index_1]

        else:
            return 1.0

    def _inflow(self, t: int) -> float:
        '''
        Compute the warfarin concentration that enters the two compartments
        is the PK/PD model.

        Arguments
        ---------
        t:
            The day for which the input is requested.

        Returns
        -------
        :
            The input value.

        Notes
        -----
        To speed up the process, total concentration is being cached for a
        number of days. For days beyond this range, concentration values are
        computed and used on each call.
        '''
        try:
            Cs = self._total_cs[t]
        except IndexError:
            Cs = sum(v.Cs(t)
                     for v in self._dose_records.values())

        Cs_gamma = (Cs * self._err(t, t > 0)) ** self._gamma
        inflow = 1 - self._E_max * \
            Cs_gamma / (self._EC_50_gamma + Cs_gamma)

        return inflow

    def _INR(self, A: List[float], t: int) -> float:
        '''
        Compute the warfarin concentration that enters the two compartments
        is the PK/PD model.

        Arguments
        ---------
        t:
            The day for which the input is requested.

        Returns
        -------
        :
            The input value.

        Notes
        -----
        To speed up the process, total concentration is being cached for a
        number of days. For days beyond this range, concentration values are
        computed and used on each call.
        '''
        INR_max = 20
        baseINR = 1

        return (baseINR +
                (INR_max*(1-A[6]*A[8]) ** self._lambda)
                ) * self._exp_e_INR(t)



# -*- coding: utf-8 -*-
'''
Buffer class
============

The base class for all buffers in `reil`.
'''

from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

import numpy as np
from reil import reilbase
from typing_extensions import Literal

T = TypeVar('T')
PickModes = Literal['all', 'random', 'recent', 'old']


class Buffer(reilbase.ReilBase, Generic[T]):
    '''
    The base class for all buffers in `reil`.
    '''

    _buffer_size = None
    _buffer_names = None
    _pick_mode = None
    _buffer_index = -1
    _count = 0

    def __init__(self,
                 buffer_size: Optional[int] = None,
                 buffer_names: Optional[List[str]] = None,
                 pick_mode: Optional[PickModes] = None) -> None:
        '''
        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        buffer_names:
            A list containing the names of buffer queues.

        pick_mode:
            The default mode to pick items from the list.
        '''
        self.setup(buffer_size, buffer_names, pick_mode)

    def setup(self,
              buffer_size: Optional[int] = None,
              buffer_names: Optional[List[str]] = None,
              pick_mode: Optional[PickModes] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        buffer_names:
            A list containing the names of buffer elements.

        pick_mode:
            The default mode to pick items from the list.

        Raises
        ------
        ValueError:
            Cannot modify `buffer_size`. The value is already set.

        ValueError:
            Cannot modify `buffer_names`. The value is already set.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        if buffer_size is not None:
            if self._buffer_size is not None:
                raise ValueError(
                    'Cannot modify buffer_size. The value is already set.')
            else:
                self._buffer_size = buffer_size

        if buffer_names is not None:
            if self._buffer_names is not None:
                raise ValueError(
                    'Cannot modify buffer_names. The value is already set.')
            else:
                self._buffer_names = buffer_names

        if self._buffer_size is not None and self._buffer_names is not None:
            self._buffer = dict((name, [0.0]*self._buffer_size)
                                for name in self._buffer_names)
        else:
            self._buffer = None

        if pick_mode is not None:
            self._pick_mode = pick_mode

        self.reset()

    def add(self, data: Dict[str, Any]) -> None:
        '''
        Append a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.

        Notes
        -----
        This implementation of `add` does not check if the buffer is full
        or if the provided names exist in the buffer queues. As a result, this
        situations will result in exceptions by the system.
        '''
        if self._buffer is None:
            raise RuntimeError('Buffer is not set up!')

        self._buffer_index += 1
        for key, v in data.items():
            self._buffer[key][self._buffer_index] = v

        self._count += 1

    def pick(self,
             count: Optional[int] = None,
             mode: Optional[PickModes] = None) -> Dict[str, Tuple[Any, ...]]:
        '''
        Return items from the buffer.

        Arguments
        ---------
        count:
            The number of items to return. If omitted, the number of items in
            the buffer is used. `count` will be ignored if `mode` is 'all'.

        mode:
            How to pick items. If omitted, the default `pick_mode` specified
            during initialization or setup is used.

        Returns
        -------
        :
            A dictionary with buffer names as keys and picked items as values.

        Raises
        ------
        ValueError:
            The number of items requested is greater than the number of items
            in the buffer.

        ValueError:
            `mode` is not one of 'all', 'random', 'recent' and 'old'.
        '''
        _mode = mode.lower() if mode is not None else self._pick_mode
        _count = count if count is not None else self._count

        if _count > self._count:
            raise ValueError('Not enough data in the buffer.')

        if _mode == 'old':
            return self._pick_old(_count)
        elif _mode == 'recent':
            return self._pick_recent(_count)
        elif _mode == 'random':
            return self._pick_random(_count)
        elif _mode == 'all':
            return self._pick_all()
        else:
            raise ValueError(
                'mode should be one of all, old, recent, or random.')

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        self._buffer_index = -1
        self._count = 0

    def _pick_old(self, count: int) -> Dict[str, Tuple[Any, ...]]:
        '''
        Return the oldest items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` oldest items
            as values.
        '''
        s = slice(count)
        return dict((name, tuple(buffer[s]))
                    for name, buffer in self._buffer.items())

    def _pick_recent(self, count: int) -> Dict[str, Tuple[Any, ...]]:
        '''
        Return the most recent items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` most recent
            items as values.
        '''
        s = slice(self._buffer_index - count + 1, self._buffer_index + 1)
        return dict((name, tuple(buffer[s]))
                    for name, buffer in self._buffer.items())

    def _pick_random(self, count: int) -> Dict[str, Tuple[Any, ...]]:
        '''
        Return a random sample of items in the buffer.

        Arguments
        ---------
        count:
            The number of items to return.

        Returns
        -------
        :
            A dictionary with buffer names as keys and `count` randomly picked
            items as values.
        '''
        index = np.random.choice(
            self._count, count, replace=False)
        return dict((name, tuple(buffer[i] for i in index))
                    for name, buffer in self._buffer.items())

    def _pick_all(self) -> Dict[str, Tuple[Any, ...]]:
        '''
        Return all items in the buffer.

        Returns
        -------
        :
            A dictionary with buffer names as keys and all items as values.
        '''
        s = slice(self._buffer_index + 1)
        return dict((name, tuple(buffer[s]))
                    for name, buffer in self._buffer.items())



# -*- coding: utf-8 -*-
'''
Feature class
=============

A datatype that accepts initial value and feature generator, and generates
new values.
'''
from __future__ import annotations

import dataclasses
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar, Union

FeatureType = TypeVar('FeatureType')


@dataclasses.dataclass
class Feature(Generic[FeatureType]):
    '''
    A datatype that accepts initial value and feature generator, and generates
    new values.

    Attributes
    ----------
    is_numerical:
        Is the feature numerical?

    value:
        The currect value of the feature.

    randomized:
        Whether the generator should produce random values.

    generator:
        A function that accepts feature characteristics and generates a new
        value

    lower:
        The lower bound for numerical features.

    upper:
        The upper bound for numerical features.

    mean:
        Mean of the distribution for numerical features.

    stdev:
        Standard Deviation of the distribution for numerical features.

    categories:
        A list of possible values for categorical features.

    probabilities:
        A list of probabilities corresponding to each possible value
        for categorical features.
    '''
    is_numerical: Optional[bool] = True
    value: Optional[Any] = None
    randomized: Optional[bool] = True
    generator: Optional[Callable[["Feature"], Any]] = lambda x: x.value

    def __post_init__(self):
        if self.is_numerical:
            self.lower: Optional[float] = None
            self.upper: Optional[float] = None
            self.mean: Optional[float] = None
            self.stdev: Optional[float] = None
        else:
            self.categories: Optional[Tuple[Any, ...]] = None
            self.probabilities: Optional[Tuple[float, ...]] = None

    @classmethod
    def categorical(cls,
                    value: Optional[Any] = None,
                    categories: Optional[Tuple[Any, ...]] = None,
                    probabilities: Optional[Tuple[float, ...]] = None,
                    randomized: Optional[bool] = None,
                    generator: Optional[Callable[[Feature], Any]] = None):
        '''
        Create a categorical Feature.

        Arguments
        ---------
        value:
            The initial value of the feature.

        randomized:
            Whether the generator should produce random values.

        generator:
            A function that gets feature characteristics and generates a new
            value

        categories:
            A list of possible values.

        probabilities:
            A list of probabilities corresponding to each possible value.
        '''
        instance = cls(is_numerical=False,
                       value=value,
                       randomized=randomized,
                       generator=generator)

        instance.categories = categories
        instance.probabilities = probabilities

        instance._categorical_validator()

        return instance

    @classmethod
    def numerical(cls,
                  value: Optional[Union[int, float]] = None,
                  lower: Optional[Union[int, float]] = None,
                  upper: Optional[Union[int, float]] = None,
                  mean: Optional[Union[int, float]] = None,
                  stdev: Optional[Union[int, float]] = None,
                  generator: Optional[Callable[[Feature], Any]] = None,
                  randomized: Optional[bool] = None):
        '''
        Create a numerical Feature.

        Arguments
        ---------
        value:
            The currect value of the feature.

        randomized:
            Whether the generator should produce random values.

        generator:
            A function that gets feature characteristics and generates a new
            value

        lower:
            The lower bound.

        upper:
            The upper bound.

        mean:
            Mean of the distribution.

        stdev:
            Standard Deviation of the distribution.
        '''
        instance = cls(is_numerical=True,
                       value=value,
                       generator=generator,
                       randomized=randomized)

        instance.lower = lower
        instance.upper = upper
        instance.mean = mean
        instance.stdev = stdev

        instance._numerical_validator()

        return instance

    def _numerical_validator(self) -> None:
        '''
        Check if the value is in the defined range.
        '''
        if self.value is not None:
            if self.lower is not None and self.value < self.lower:
                raise ValueError(
                    f'value={self.value} is less than lower={self.lower}.')

            if self.upper is not None and self.value > self.upper:
                raise ValueError(
                    f'value={self.value} is greater than upper={self.upper}.')

    def _categorical_validator(self) -> None:
        '''
        Check if the value is in the defined categories and probabilities add
        up to one.
        '''
        if self.value is not None:
            if (self.categories is not None and
                    self.value not in self.categories):
                raise ValueError(
                    f'value={self.value} is in '
                    f'the categories={self.categories}.')

        if self.probabilities is not None:
            if abs(sum(self.probabilities) - 1.0) > 1e-6:
                raise ValueError('probabilities should add up to 1.0.'
                                 f'Got {sum(self.probabilities)}')
            if self.categories is None:
                raise ValueError(
                    'probabilities cannot be set for None categories.')
            if len(self.probabilities) != len(self.categories):
                raise ValueError('Size mismatch. '
                                 f'{len(self.categories)} categories vs. '
                                 f'{len(self.categories)} probabilities')

    def generate(self) -> None:
        '''
        Generate a new value using the generator.
        '''
        if self.generator is None:
            return

        if self.randomized or self.value is None:
            self.value = self.generator(self)



# -*- coding: utf-8 -*-
'''
experiment class
=================

This `environment` class provides an experimentation environment for any
reinforcement learning agent on any subject. 


'''

from math import log10, ceil
from pathlib import Path
# from zipfile import ZipFile 
import os
import pandas as pd
import numpy as np

from reil.environments import Environment
import reil.agents as agents
import reil.subjects as subjects


class Experiment(Environment):
    '''
    Provide an experimentation environment for agents and subjects.

    Attributes
    ----------

    Methods
    -------
        add: add a set of objects (agents/ subjects) to the environment.
        remove: remove objects (agents/ subjects) from the environment.
        assign: assign agents to subjects.
        run: run the experiment using agents on subjects.
        load: load an object (agent/ subject) or an environment.
        save: save an object (agent/ subject) or the current environment.

    Agents act on subjects and receive the reward of their action and the new state of subjects.
    '''

    def __init__(self, **kwargs):
        '''
        Create a new experiment.

        Arguments
        ---------
            filename: if given, Experiment attempts to open a saved experiment by openning the file. This argument cannot be used along other arguments.
            path: path of the file to be loaded (should be used with filename).
            number_of_subjects: number of subjects to create for experimentation (Default = 1)
            max_steps: maximum number of steps to go for each subject (Default = 10,000)
            save_subjects: whether to save subject files or not (Default = True)
        '''
        Environment.__init__(self, **kwargs)
        if 'filename' in kwargs:
            self.load(path=kwargs.get('path', '.'),
                      filename=kwargs['filename'])
            return

        self._agent = {}
        self._subject = {}
        self._assignment_list = {}
        self._number_of_subjects = 1
        self._max_steps = 10000
        self._save_subjects = True
        self._file_index_generator = None
        self.set_params(**kwargs)

    def _default_file_index_generator(self, number_of_subjects):
        digits = ceil(log10(number_of_subjects))
        for subject_ID in range(number_of_subjects):
            print(subject_ID)
            yield str(subject_ID).rjust(digits, '0')

    def generate_subjects(self, **kwargs):
        number_of_subjects = kwargs.get('number_of_subjects', self._number_of_subjects)
        subjects_path = kwargs.get('subjects_path', './subjects')
        file_index_generator = kwargs.get('file_index_generator',
            self._file_index_generator if self._file_index_generator is not None else self._default_file_index_generator)
        print(file_index_generator)
        self._number_of_subjects = number_of_subjects
        # save_subjects = kwargs.get('save_subjects', self._save_subjects)

        for subject_name, subject in self._subject.items():
            subject.reset()
            print('generating')
            for file_index in file_index_generator(number_of_subjects):
                print(file_index)
                filename = subject_name + file_index
                if os.path.exists(os.path.join(subjects_path, filename + '.pkl')):
                    print(f'{filename} already exists! Skipping the file!')
                else:
                    subject.save(filename=filename, path=subjects_path)
                    subject.reset()
                # with ZipFile(subject_name,'w') as zip:
                #     for file in file_paths: 
                #         zip.write(file)

    def run(self, **kwargs):
        '''
        Run the experiment.

        Arguments
        ---------
            max_steps: maximum number of steps in the epoch (Default = 10,000)

        '''
        max_steps = kwargs.get('max_steps', self._max_steps)
        number_of_subjects = kwargs.get('number_of_subjects', self._number_of_subjects)
        subjects_path = kwargs.get('subjects_path', './subjects')
        outputs_path = kwargs.get('outputs_path', './' + '/outputs')
        file_index_generator = kwargs.get('file_index_generator',
            self._file_index_generator if self._file_index_generator is not None else self._default_file_index_generator)

        output_dir = Path(outputs_path)
        output_dir.mkdir(parents=True, exist_ok=True)

        for agent_name, agent in self._agent.items():
            print(f'Agent: {agent_name}')
            agent.training_mode = False

            try:
                subject_name, _id = self._assignment_list[agent_name]
            except KeyError:
                continue

            subject = self._subject[subject_name]

            for file_index in file_index_generator(number_of_subjects):
                print(file_index)
                history = pd.DataFrame(columns=['state', 'action', 'q', 'reward'])

                filename = subject_name + file_index
                print(f'Subject: {filename}')
                temp_agent_list = subject._agent_list
                # THIs SHOULD BE subject.load(...), I ADDED ._patient TEMPORARILY! ALSO,
                # DELETE subject.reset()
                subject.reset()
                subject._patient.load(filename=filename, path=subjects_path)  # './' + subject_name)
                subject._agent_list = temp_agent_list

                steps = 0
                while not subject.is_terminated and steps < max_steps:
                    steps += 1

                    state = subject.state
                    possible_actions = subject.possible_actions
                    action = agent.act(state, actions=possible_actions)
                    try:
                        q = agent._q(state, action)
                    except AttributeError:
                        q = np.nan
                    reward = subject.take_effect(action, _id)

                    history.loc[len(history.index)] = [state, action, q, reward]

                if subject.is_terminated:
                    for affected_agent in self._agent.keys():
                        try:
                            if (self._assignment_list[affected_agent][0] == subject_name) & \
                                    (affected_agent != agent_name):
                                history[-1] = -reward
                        except KeyError:
                            pass

                history.to_pickle(output_dir / (agent_name + '@' + filename + '.pkl'))

    def __repr__(self):
        try:
            return 'Exp: \n Agents:\n' + \
                '\n\t'.join((a.__repr__() for a in self._agent.values())) + \
                '\nSubjects:\n' + \
                '\n\t'.join((s.__repr__() for s in self._subject.values()))
        except AttributeError:
            return 'Experiment: New'


if __name__ == "__main__":
    from reil.agents import RandomAgent
    from reil.subjects import WarfarinModel_v4

    def index_generator(number_of_subjects, start_number=3, digits=5):
        for i in range(start_number, start_number + number_of_subjects):
            yield str(i).rjust(digits,'0')

    exp = Experiment(agent={'R': RandomAgent()},
                     subject={'': WarfarinModel_v4()},
                     assignment_list={'R': ('', 1)})

    exp.generate_subjects(number_of_subjects=7, subjects_path='./patients', file_index_generator=index_generator)
    exp.run(subjects_path='./patients', outputs_path='./experiment_outputs', file_index_generator=index_generator)



# -*- coding: utf-8 -*-
'''
Environment class
==================

The base class of all learning environments in which one or more `agents` act
on one or more `subjects`.
'''
import inspect
import pathlib
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Tuple, Union, cast

from reil import agents as rlagents
from reil import stateful
from reil import subjects as rlsubjects
from reil.datatypes import InteractionProtocol, ReilData
from reil.utils import instance_generator as rlgenerator

AgentSubjectTuple = Tuple[str, str]
Entity = Union[rlagents.Agent, rlsubjects.Subject]
EntityGenerator = Union[rlgenerator.InstanceGenerator[rlagents.Agent],
                        rlgenerator.InstanceGenerator[rlsubjects.Subject]]


class Environment(stateful.Stateful):
    '''
    Provide an interaction and learning environment for `agents` and
    `subjects`.

    Notes
    -----
    `Agents` act on `subjects` and receive the reward of their action and
    the new state of those `subjects`. Then `agents` learn based on this
    information to improve their actions.

    `Pass`: visiting all protocols in the interaction sequence, once.

    `Epoch`: For each subject, an epoch is one time reaching its terminal
    state. If the subject is an instance generator, then the generator should
    reach to terminal state, not just its current instance.
    '''

    def __init__(self,
                 entity_dict: Optional[
                     Dict[str, Union[Entity, EntityGenerator, str]]] = None,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        entity_dict:
            a dictionary that contains `agents`, `subjects`, and
            `generators`.
        '''
        super().__init__(name=kwargs.get('name', __name__),
                         logger_name=kwargs.get('logger_name', __name__),
                         **kwargs)

        self._agents: Dict[str, rlagents.AgentType] = {}
        self._subjects: Dict[str, rlsubjects.SubjectType] = {}
        self._instance_generators: Dict[str, EntityGenerator] = {}
        self._assignment_list: Dict[
            AgentSubjectTuple,
            Tuple[Union[int, None], Union[int, None]]] = \
            defaultdict(lambda: (None, None))
        self._epochs: Dict[str, int] = defaultdict(int)
        self._agent_observers: Dict[
            Tuple[str, str], Generator[Union[ReilData, None], Any, None]] = {}

        if entity_dict is not None:
            self.add(entity_dict)

    def add(self,
            entity_dict: Dict[str, Union[Entity, EntityGenerator, str]]
            ) -> None:
        '''
        Add `agents` and `subjects` to the environment.

        Arguments
        ---------
        entity_dict:
            a dictionary consist of `agent`/ `subject` name and the
            respective entity. Names should be unique, otherwise overwritten.
            To assign one entity to multiple names, use the name in the
            first assignment as the value of the dict for other assignments.
            For example:
            >>> env.add({'agent_1': Agent(), 'agent_2': 'agent_1'})

            When using name as value, the name is being looked up first in
            instance generators, then agents, and finally subjects. Whichever
            contains the name first, the entity corresponding to that instance
            is being used.

        Notes
        -----
        Reusing an `InstanceGenerator` produces unintended consequences.

        Raises
        ------
        ValueError
            An entity is being reused without being defined first.

        TypeError
            The entity is niether an `agent` [generator] nor a `subject`
            [generator].
        '''
        for name, obj in entity_dict.items():
            if isinstance(obj, str):
                _obj = self._instance_generators.get(
                    obj, self._agents.get(
                        obj, self._subjects.get(obj)))
                if _obj is None:
                    raise ValueError(f'entity {obj} defined for {name} is '
                                     'not in the list of agents, subjects, '
                                     'and generators.')
            else:
                _obj = obj
            if isinstance(_obj, rlgenerator.InstanceGenerator):
                self._instance_generators.update({name: _obj})
            elif isinstance(_obj, rlagents.NoLearnAgent):
                self._agents.update({name: _obj})
            elif isinstance(_obj, rlsubjects.Subject):
                self._subjects.update({name: _obj})
            else:
                raise TypeError(
                    f'entity {name} is niether an agent nor a subject.')

        for name, generator in self._instance_generators.items():
            _, obj = next(generator)
            if isinstance(obj, rlagents.Agent):
                self._agents.update({name: obj})
            elif isinstance(obj, rlsubjects.Subject):
                self._subjects.update({name: obj})
            else:
                raise TypeError(
                    f'entity {name} is niether an agent nor a subject.')

    def remove(self, entity_names: Tuple[str, ...]) -> None:
        '''
        Remove `agents`, `subjects`, or `instance_generators` from
        the environment.

        Arguments
        ---------
        entity_names:
            A list of `agent`/ `subject` names to be deleted.

        Notes
        -----
        This method removes the item from both `agents` and `subjects`
        lists. Hence, it is not recommended to use the same name for both
        an `agent` and a `subject`.
        '''
        for name in entity_names:
            if name in self._agents:
                del self._agents[name]
            if name in self._subjects:
                del self._subjects[name]
            if name in self._instance_generators:
                del self._instance_generators[name]

    def simulate_one_pass(self) -> None:
        '''
        Go through the interaction sequence for one pass and
        simulate interactions accordingly.
        '''
        raise NotImplementedError

    def simulate_passes(self, passes: int) -> None:
        '''
        Go through the interaction sequence for a number of passes and
        simulate interactions accordingly.

        Arguments
        ---------
        passes:
            The number of passes that simulation should go.
        '''
        for _ in range(passes):
            self.simulate_one_pass()

    def simulate_to_termination(self) -> None:
        '''
        Go through the interaction sequence and simulate interactions
        accordingly, until all `subjects` are terminated.

        Notes
        -----
        To avoid possible infinite loops caused by normal `subjects`,
        this method is only available if all `subjects` are generated
        by `instance generators`.

        Raises
        ------
        TypeError:
            Attempt to call this method will normal subjects in the interaction
            sequence.
        '''
        raise NotImplementedError

    @staticmethod
    def interact_once(
            agent_id: int,
            agent_observer: Generator[Union[ReilData, None], Any, None],
            subject_instance: rlsubjects.Subject,
            state_name: str,
            reward_function_name: str,
            epoch: int) -> None:
        '''
        Allow `agent` and `subject` to interact once.

        Attributes
        ----------
        agent_id:
            Agent's ID by which it is registered at the `subject`.

        subject_id:
            Subject's ID by which it is registered at the `agent`.

        agent_instance:
            An instance of an `agent` that takes the action.

        subject_instance:
            An instance of a `subject` that computes reward, determines
            possible actions, and takes the action.

        state_name:
            A string that specifies the state definition.

        reward_function_name:
            A string that specifies the reward function definition.

        epoch:
            The epoch of of the current run. This value is used by the `agent`
            to determine the action.

        Returns
        -------
        :
            Reward received by `subject`  and state before taking an action and
            the action that `agent` took.

        Notes
        -----
        This method does not check whether the `subject` is terminated.

        If no possible actions are available, `None` will be returned for
        action.
        '''
        agent_observer.send(subject_instance.reward(
            name=reward_function_name, _id=agent_id))

        state = subject_instance.state(name=state_name, _id=agent_id)
        possible_actions = subject_instance.possible_actions(agent_id)
        if possible_actions:
            action = agent_observer.send({'state': state,
                                          'actions': possible_actions,
                                          'epoch': epoch})
            subject_instance.take_effect(cast(ReilData, action), agent_id)

    @classmethod
    def interact_n_times(
            cls,
            agent_id: int,
            agent_observer: Generator[Union[ReilData, None], Any, None],
            subject_instance: rlsubjects.Subject,
            state_name: str,
            reward_function_name: str,
            epoch: int,
            times: int = 1) -> None:
        '''
        Allow `agent` and `subject` to interact at most `times` times.

        Attributes
        ----------
        agent_id:
            Agent's ID by which it is registered at the subject.

        subject_id:
            Subject's ID by which it is registered at the `agent`.

        agent_instance:
            An instance of an `agent` that takes the action.

        subject_instance:
            An instance of a `subject` that computes reward, determines
            possible actions, and takes the action.

        state_name:
            A string that specifies the state definition.

        reward_function_name:
            A string that specifies the reward function definition.

        epoch:
            The epoch of of the current run. This value is used by the `agent`
            to determine the action.

        times:
            The number of times the `agent` and the `subject` should interact.

        Returns
        -------
        :
            A list of subject's reward and state before taking an action
            and agent's action.

        Notes
        -----
        If subject is terminated before "times" iterations, the result will
        be truncated and returned. In other words, the output will not
        necessarily have a lenght of "times".
        '''
        for _ in range(times):
            cls.interact_once(agent_id, agent_observer, subject_instance,
                              state_name, reward_function_name, epoch)

    @classmethod
    def interact_while(
            cls,
            agent_id: int,
            agent_observer: Generator[Union[ReilData, None], Any, None],
            subject_instance: rlsubjects.Subject,
            state_name: str,
            reward_function_name: str,
            epoch: int) -> None:
        '''
        Allow `agent` and `subject` to interact until `subject` is terminated.

        Attributes
        ----------
        agent_id:
            Agent's ID by which it is registered at the subject.

        agent_instance:
            An instance of an agent that takes the action.

        subject_instance:
            An instance of a subject that computes reward,
            determines possible actions, and takes the action.

        state_name:
            A string that specifies the state definition.

        reward_function_name:
            A string that specifies the reward function definition.

        epoch:
            The epoch of of the current run. This value is used by the agent
            to determine the action.

        Returns
        -------
        :
            A list of subject's reward and state before taking an action and
            agent's action.

        Notes
        -----
        For `instance generators`, only the current
        instance is run to termination, not the whole generator.
        '''
        while not subject_instance.is_terminated(agent_id):
            cls.interact_once(agent_id, agent_observer, subject_instance,
                              state_name, reward_function_name, epoch)

    def assert_protocol(self, protocol: InteractionProtocol) -> None:
        '''
        Check whether the given protocol:

        * contains only entities that are known to the `environment`.

        * unit is one of the possible values.

        Arguments
        ---------
        protocol:
            An interaction protocol.

        Raises
        ------
        ValueError
            `agent` or `subject` is not defined.

        ValueError
            `unit` is not one of `interaction`, `instance`, or `epoch`.
        '''
        if protocol.agent.name not in self._agents:
            raise ValueError(f'Unknown agent name: {protocol.agent.name}.')
        if protocol.subject.name not in self._subjects:
            raise ValueError(f'Unknown subject name: {protocol.subject.name}.')
        if protocol.unit not in ('interaction', 'instance', 'epoch'):
            raise ValueError(
                f'Unknown unit: {protocol.unit}. '
                'It should be one of interaction, instance, or epoch. '
                'For subjects of non-instance generator, epoch and '
                'instance are equivalent.')
        # if (protocol.agent_name in self._instance_generators or
        #     protocol.subject_name in self._instance_generators) and

    def register(self,
                 interaction_protocol: InteractionProtocol,
                 get_agent_observer: bool = False) -> None:
        '''
        Register the `agent` and `subject` of an interaction protocol.

        Arguments
        ---------
        interaction_protocol:
            The protocol whose `agent` and `subject` should be registered.

        get_agent_observer:
            If `True`, the method calls the `observe` method of the `agent`
            with `subject_id`, and adds the resulting generator to the list
            of observers.

        Notes
        -----
        When registration happens for the first time, agents and subjects
        get any ID that the counterpart provides. However, in the follow up
        registrations, `entities` attempt to register with the same ID to
        have access to the same information.
        '''
        a_name = interaction_protocol.agent.name
        a_stat = interaction_protocol.agent.statistic_name
        s_name = interaction_protocol.subject.name
        a_s_name = (a_name, s_name)

        a_id, s_id = self._assignment_list[a_s_name]
        a_id = self._subjects[s_name].register(entity_name=a_name, _id=a_id)
        s_id = self._agents[a_name].register(entity_name=s_name, _id=s_id)

        self._assignment_list[a_s_name] = (a_id, s_id)

        if get_agent_observer:
            self._agent_observers[a_s_name] = \
                self._agents[a_name].observe(s_id, a_stat)

    def close_agent_observer(self, protocol: InteractionProtocol) -> None:
        '''
        Close an `agent_observer` corresponding to `protocol`.

        Before closing the observer, the final `reward` and `state` of the
        system are passed on to the observer.

        Attributes
        -----------
        protocol:
            The protocol whose `agent_observer` should be closed.

        Notes
        -----
        This method should only be used if a `subject` is terminated.
        Otherwise, the `agent_observer` might be expecting to receive different
        values, and it will corrupt the training data for the `agent`.
        '''
        agent_name = protocol.agent.name
        subject_name = protocol.subject.name
        r_func_name = protocol.reward_function_name
        state_name = protocol.state_name
        a_s_names = (agent_name, subject_name)

        if inspect.getgeneratorstate(
                self._agent_observers[a_s_names]) != inspect.GEN_SUSPENDED:
            return

        a_id, _ = cast(Tuple[int, int],
                       self._assignment_list[a_s_names])
        reward = self._subjects[subject_name].reward(
            name=r_func_name, _id=a_id)
        state = self._subjects[subject_name].state(
            name=state_name, _id=a_id)

        self._agent_observers[a_s_names].send(reward)
        self._agent_observers[a_s_names].send({'state': state,
                                               'actions': None,
                                               'epoch': None})
        self._agent_observers[a_s_names].close()

    def reset_subject(self, subject_name: str) -> bool:
        '''
        When a `subject` is terminated for all interacting `agents`, this
        function is called to reset the `subject`.

        If the `subject` is an `InstanceGenerator`, a new instance is created.
        If reset is successful, `epoch` is incremented by one.

        Attributes
        ----------
        subject_name:
            Name of the `subject` that is terminated.

        Returns
        -------
        :
            `True` if the `instance_generator` for the `subject` is still
            active, `False` if it hit `StopIteration`.

        Notes
        -----
        `Environment.reset_subject` only resets the `subject`. It does not
        get the statistics for that `subject`.
        '''
        if subject_name in self._instance_generators:
            # get a new instance if possible,
            # if not instance generator returns StopIteration.
            # So, increment epoch by 1, then if the generator is not
            # terminated, get a new instance.
            # If the generator is terminated, check if it is finite. If
            # infinite, call it again to get a subject. If not, disable reward
            # for the current subject, so that agent_observer does not raise
            # exception.
            try:
                _, self._subjects[subject_name] = cast(
                    Tuple[int, rlsubjects.SubjectType],
                    next(self._instance_generators[subject_name]))

            except StopIteration:
                self._epochs[subject_name] += 1
                if self._instance_generators[subject_name].is_terminated():
                    self._subjects[subject_name].reward.disable()
                    # if self._instance_generators[subject_name].is_finite:
                    #     self._subjects[subject_name].reward.disable()
                    # else:
                    #     _, self._subjects[subject_name] = cast(
                    #         Tuple[int, rlsubjects.SubjectType],
                    #         next(self._instance_generators[subject_name]))
                else:
                    _, self._subjects[subject_name] = cast(
                        Tuple[int, rlsubjects.SubjectType],
                        next(self._instance_generators[subject_name]))
                return False
        else:
            self._epochs[subject_name] += 1
            self._subjects[subject_name].reset()

        return True

    def load(self,
             entity_name: Union[List[str], str] = 'all',
             filename: Optional[str] = None,
             path: Optional[Union[pathlib.Path, str]] = None) -> None:
        '''
        Load an entity or an `environment` from a file.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        entity_name:
            If specified, that entity (`agent` or `subject`) is being
            loaded from file. 'all' loads an `environment`.

        Raises
        ------
        ValueError
            The filename is not specified.
        '''
        _filename: str = filename or self._name
        _path = pathlib.Path(path if path is not None else self._path)

        if entity_name == 'all':
            super().load(filename=_filename, path=_path)
            self._agents: Dict[str, rlagents.AgentType] = {}
            self._subjects: Dict[str, rlsubjects.SubjectType] = {}
            for name, obj_type in self._env_data['agents']:
                self._agents[name] = obj_type.from_pickle(
                    path=(_path / f'{_filename}.data'), filename=name)
            for name, obj_type in self._env_data['subjects']:
                self._subjects[name] = obj_type.from_pickle(
                    path=(_path / f'{_filename}.data'), filename=name)

            del self._env_data

        else:
            for obj in entity_name:
                if obj in self._agents:
                    self._agents[obj].load(
                        path=(_path / f'{_filename}.data'), filename=obj)
                    self._agents[obj].reset()
                elif obj in self._subjects:
                    self._subjects[obj].load(
                        path=(_path / f'{_filename}.data'), filename=obj)
                    self._subjects[obj].reset()

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[pathlib.Path, str]] = None,
             data_to_save: Union[List[str], str] = 'all'
             ) -> Tuple[pathlib.Path, str]:
        '''
        Save an entity or the `environment` to a file.

        Arguments
        ---------
        filename:
            The name of the file to be saved.

        path:
            The path of the file to be saved.

        entity_name:
            If specified, that entity (`agent` or `subject`) is being saved
            to file. 'all' saves the `environment`.

        Raises
        ------
        ValueError
            The filename is not specified.
        '''
        _filename = filename or self._name
        _path = pathlib.Path(path or self._path)

        if data_to_save == 'all':
            self._env_data: Dict[str, List[Any]] = defaultdict(list)

            for name, agent in self._agents.items():
                _, fn = agent.save(
                    path=_path / f'{_filename}.data', filename=name)
                self._env_data['agents'].append((fn, type(agent)))

            for name, subject in self._subjects.items():
                _, fn = subject.save(
                    path=_path / f'{_filename}.data', filename=name)
                self._env_data['subjects'].append((fn, type(subject)))

            super().save(
                filename=_filename, path=_path,
                data_to_save=tuple(v for v in self.__dict__
                                   if v not in ('_agents', '_subjects')))

            del self._env_data
        else:
            for obj in data_to_save:
                if obj in self._agents:
                    self._agents[obj].save(
                        path=_path / f'{_filename}.data', filename=obj)
                elif obj in self._subjects:
                    self._subjects[obj].save(
                        path=_path / f'{_filename}.data', filename=obj)

        return _path, _filename

    def report_statistics(self):
        raise NotImplementedError

    def __repr__(self) -> str:
        try:
            return super().__repr__() + '\n Agents:\n' + \
                '\n\t'.join((a.__repr__() for a in self._agents.values())) + \
                '\nSubjects:\n' + \
                '\n\t'.join((s.__repr__() for s in self._subjects.values()))
        except AttributeError:
            return super().__repr__()



# from __future__ import annotations

# import copy
# import itertools
# import operator
# from collections.abc import MutableSequence
# from numbers import Number
# from typing import (Any, Callable, Dict, Generator, Iterable, List,
#                     Mapping, Optional, Sequence, Tuple, TypeVar, Union)


# T = TypeVar('T')  # Type for value
# C = TypeVar('C')  # Type for categories
# Categories = Sequence[C]

# N = Union[Number, Sequence[Number]]  # Type for numerical value

# RLDataClass = TypeVar('RLDataClass', bound='BaseRLData')
# Normalized = Union[Number, Sequence[Number], None]
# Normalizer = Callable[[RLDataClass], Normalized]

# R = Sequence[Union[Dict[str, Any], RLDataClass]]

# class BaseRLData:
#     __slots__ = ['name', '_value', '_categorical',
#                  '_normalized', '_normalizer', '_lazy_evaluation']

#     def __init__(self,
#                  name: str,
#                  value: T,
#                  categorical: bool,
#                  normalizer: Normalizer,
#                  lazy_evaluation: bool):
#         self.name = name
#         self._value = value
#         self._categorical = categorical
#         self._normalizer = normalizer
#         self._lazy_evaluation = lazy_evaluation
#         self._normalize()

#     def _normalize(self) -> None:
#         if self._lazy_evaluation:
#             self._normalized = None
#         else:
#             self._normalized = self._normalizer(self)

#     @property
#     def value(self) -> T:
#         return self._value

#     @value.setter
#     def value(self, v: T):
#         self._value = v
#         self._normalize()

#     @property
#     def categorical(self) -> bool:
#         return self._categorical

#     @property
#     def normalizer(self) -> Normalizer:
#         return self._normalizer

#     @normalizer.setter
#     def normalizer(self, func: Normalizer):
#         if callable(func):
#             self._normalizer = func
#             self._normalize()
#         else:
#             raise TypeError('Callable argument expected.')

#     @property
#     def lazy_evaluation(self) -> bool:
#         return self._lazy_evaluation

#     @lazy_evaluation.setter
#     def lazy_evaluation(self, lazy_evaluation: bool):
#         self._lazy_evaluation = lazy_evaluation
#         self._normalize()

#     @property
#     def normalized(self) -> Normalized:
#         if self._normalized is None:
#             self._normalized = self._normalizer(self)

#         return self._normalized

#     def as_dict(self) -> Dict[str, Any]:
#         return {'name': self.name, 'value': self.value}

#     def __eq__(self, other: BaseRLData) -> bool:
#         return (isinstance(other, type(self)) and
#                 all((self.name == other.name,
#                      self._value == other._value,
#                      self._categorical == other._categorical,
#                      self._normalizer == other._normalizer,
#                      self._lazy_evaluation == other._lazy_evaluation)))

#     def __hash__(self) -> int:
#         return hash((self.name,
#                      self._value,
#                      self._categorical,
#                      self._normalizer,
#                      self._lazy_evaluation))

#     def __str__(self) -> str:
#         return f'{{{self.name}: {self.value}}}'


# class CategoricalData(BaseRLData):
#     __slots__ = ['_categories']

#     def __init__(self,
#                  name: str,
#                  value: T,
#                  categories: Optional[Categories] = None,
#                  normalizer: Optional[Normalizer] = None,
#                  lazy_evaluation: bool = False):

#         super().__init__(name=name,
#                          value=value,
#                          categorical=True,
#                          normalizer=normalizer if normalizer is not None else self._default_normalizer,
#                          lazy_evaluation=True)  # Do not evaluate before assigning categories.

#         self._categories = categories
#         self.lazy_evaluation = lazy_evaluation

#     @staticmethod
#     def _default_normalizer(x: CategoricalData) -> Normalized:
#         if x.categories is None:  # categories are not defined
#             return None

#         if isinstance(x.value, type(x.categories[0])):
#             return list(int(x_i == x.value) for x_i in x.categories)  # type: ignore

#         # No need for the following type checking, since it was done in object creation.
#         # elif isinstance(x.value[0], type(x.categories[0])):
#         #     return list(int(x_i == v) for v in x.value for x_i in x.categories)  # type: ignore

#         return list(int(x_i == v) for v in x.value for x_i in x.categories)  # type: ignore

#     @property
#     def categories(self) -> Union[Categories, None]:
#         return self._categories

#     @categories.setter
#     def categories(self, cat: Union[Categories, None]):
#         val = self.value  # Added to speed up the code
#         if cat is None or val in (None, (), []):
#             self._categories = cat
#             self._normalized = None
#         elif not isinstance(cat, (list, tuple)):
#             raise TypeError(
#                 f'A sequence (list or tuple) was expected. Received a {type(cat)}.\n{self.__repr__()}')
#         elif CategoricalData._type_check(val, cat):
#             self._categories = cat
#             self._normalize()
#         else:
#             raise ValueError(
#                 f'Categories list {cat} does not include the current value: {self.value}.\n{self.__repr__()}')

#     @property
#     def value(self) -> T:
#         return super().value

#     @value.setter
#     def value(self, v: T):
#         if CategoricalData._type_check(v, self.categories):
#             super(CategoricalData, self.__class__).value.fset(self, v)
#         else:
#             raise ValueError(f'{v} is not in categories: {self._categories}.')

#     def as_dict(self) -> Dict[str, Any]:
#         temp_dict = super().as_dict()
#         temp_dict.update({'categories': self.categories, 'categorical': True})
#         return temp_dict

#     @staticmethod
#     def _type_check(val: Any, cat: Any):
#         return (isinstance(val, type(cat[0])) and val in cat) or \
#                 (isinstance(val[0], type(cat[0])) and all(v in cat for v in val))

#     def __eq__(self, other: CategoricalData) -> bool:
#         return super().__eq__(other) and (self._categories == other._categories)

#     def __hash__(self) -> int:
#         return super().__hash__() 
#         hash((self.name,
#                      self._value,
#                      self._categorical,
#                      self._normalizer,
#                      self._lazy_evaluation))

#     def __str__(self):
#         return f'{self.name}: {self._value} from {self._categories}'


# class NumericalData(BaseRLData):
#     __slots__ = ['_lower', '_upper']

#     def __init__(self,
#                  name: str,
#                  value: N,
#                  lower: Optional[Number] = None,
#                  upper: Optional[Number] = None,
#                  normalizer: Optional[Normalizer] = None,
#                  lazy_evaluation: bool = False):

#         super().__init__(name=name,
#                          value=value,
#                          categorical=False,
#                          normalizer=normalizer if normalizer is not None else self._default_normalizer,
#                          lazy_evaluation=True)  # Do not evaluate before assigning lower and upper.
#         self._lower = None
#         self._upper = None
#         self.lower = lower
#         self.upper = upper
#         self.lazy_evaluation = lazy_evaluation

#     @staticmethod
#     def _default_normalizer(x: NumericalData) -> Normalized:
#         try:
#             denominator = x.upper - x.lower  # type: ignore
#         except TypeError:  # upper or lower are not defined
#             return None

#         sequence_check = isinstance(x.value, (list, tuple))
#         try:
#             if sequence_check:
#                 return list((v - x.lower) / denominator for v in x.value)
#             else:
#                 return (x.value - x.lower) / denominator  # type: ignore
#         except ZeroDivisionError:
#             return [1] * len*x.value if sequence_check else 1  # type: ignore

#     @staticmethod
#     def _check_new_bound(value: N,
#                          new_bound: Number,
#                          bound_type: str = 'lower'):
#         if bound_type == 'lower':
#             op = operator.le
#             func = min
#         else:
#             op = operator.ge
#             func = max

#         try:
#             return op(new_bound, func(value))  # type: ignore
#         except TypeError:
#             return op(new_bound, value)

#     @property
#     def lower(self) -> Union[Number, None]:
#         return self._lower

#     @lower.setter
#     def lower(self, l: Union[Number, None]):
#         val = self.value  # Added to speed up the code
#         if l is None or val in (None, (), []):
#             self._lower = l
#             self._normalized = None
#         elif not isinstance(l, (int, float)):
#             raise TypeError(
#                 f'A numerical value expected. Received a {type(l)}.\n{self.__repr__()}')
#         elif self._lower is not None and self._lower >= l:
#             self._lower = l
#             self._normalize()
#         elif self._check_new_bound(val, l, 'lower'):
#             self._lower = l
#             self._normalize()
#         else:
#             raise ValueError(
#                 f'Lower bound {l} is greater than the current value: {self.value}.\n{self.__repr__()}')

#     @property
#     def upper(self) -> Union[Number, None]:
#         return self._upper

#     @upper.setter
#     def upper(self, u: Union[Number, None]):
#         val = self.value  # Added to speed up the code
#         if u is None or val in (None, (), []):
#             self._upper = u
#             self._normalized = None
#         elif not isinstance(u, (int, float)):
#             raise TypeError(
#                 f'A numerical value expected. Received a {type(u)}.\n{self.__repr__()}')
#         elif self._upper is not None and self._upper <= u:
#             self._upper = u
#             self._normalize()
#         elif self._check_new_bound(val, u, 'upper'):
#             self._upper = u
#             self._normalize()
#         else:
#             raise ValueError(
#                 f'Upper bound {u} is less than the current value: {self.value}.\n{self.__repr__()}')

#     @property
#     def value(self) -> N:
#         return super().value

#     @value.setter
#     def value(self, v: N):
#         if isinstance(v, (list, tuple)):
#             if not all(self.lower <= v_i <= self.upper for v_i in v):
#                 raise ValueError(
#                     f'{v} is not in range: [{self.lower}, {self.upper}].')
#         elif not (self.lower <= v <= self.upper):  # type: ignore
#             raise ValueError(
#                 f'{v} is not in range: [{self.lower}, {self.upper}].')
#         else:
#             super(NumericalData, self.__class__).value.fset(self, v)


#     def as_dict(self) -> Dict[str, Any]:
#         temp_dict = super().as_dict()
#         temp_dict.update(
#             {'lower': self.lower, 'upper': self.upper, 'categorical': False})
#         return temp_dict

#     def __eq__(self, other: NumericalData) -> bool:
#         return (super().__eq__(other) and
#                 (self._lower == other._lower) and
#                 (self._upper == other._upper))

#     def __str__(self):
#         return f'{self.name}: {self.value} of range [{self.lower}, {self.upper}]'


# class RangedData:
#     def __new__(cls,
#                 name: str,
#                 categorical: bool,
#                 value: Any,
#                 **kwargs) -> Union[CategoricalData, NumericalData]:
#         if categorical:
#             obj = CategoricalData(name=name, value=value,
#                                   categories=kwargs.get('categories'),
#                                   normalizer=kwargs.get('normalizer'),
#                                   lazy_evaluation=kwargs.get('lazy_evaluation', True))
#         else:
#             obj = NumericalData(name=name, value=value,
#                                 lower=kwargs.get('lower'),
#                                 upper=kwargs.get('upper'),
#                                 normalizer=kwargs.get('normalizer'),
#                                 lazy_evaluation=kwargs.get('lazy_evaluation', True))

#         return obj


# class RLData(MutableSequence):
#     def __init__(self,
#         data: Union[Mapping[str, Any], R, Generator[R, Any, Any]],
#         lazy_evaluation: Optional[bool] = None) -> None:
#         '''
#         Create an RLData instance.

#         Attributes:
#             data: data is either a sequence of dict-like objects that include 'name'
#                   or a mapping whose keys are names of variables.
#                   Other attributes are optional. If the class cannot find 'categorical',
#                   it attemps to find 'categories'. If fails, the object is assumed numerical.
#             lazy_evaluation: whether to store normalized values or compute on-demand.
#                              If not provided, class looks for 'lazy evaluation' in
#                              each object. If fails, True is assumed.
#         '''
#         self._data = []
#         self._clear_temps()

#         def _from_tuple(d: Tuple[str, Mapping[str, Any]]) -> RangedData:
#             if isinstance(d, BaseRLData):
#                 return copy.copy(d)

#             return RangedData(
#                     name=d[0],
#                     value=d[1].get('value'),
#                     # if categorical is not available, check if categories is available.
#                     categorical=d[1].get('categorical', d[1].get(
#                         'categories') is not None),
#                     **{'categories': d[1].get('categories'),
#                         'lower': d[1].get('lower'),
#                         'upper': d[1].get('upper'),
#                         'normalizer': d[1].get('normalizer'),
#                         'lazy_evaluation': lazy_evaluation if lazy_evaluation is not None
#                             else d[1].get('lazy_evaluation', True)})

#         def _from_dict(v: Union[BaseRLData, Mapping[str, Any]]) -> RangedData:
#             if isinstance(v, BaseRLData):
#                 return copy.copy(v)

#             return RangedData(
#                     name=v['name'],
#                     value=v.get('value'),
#                     # if categorical is not available, check if categories is available.
#                     categorical=v.get('categorical', v.get(
#                         'categories') is not None),
#                     **{'categories': v.get('categories'),
#                         'lower': v.get('lower'),
#                         'upper': v.get('upper'),
#                         'normalizer': v.get('normalizer'),
#                         'lazy_evaluation': lazy_evaluation if lazy_evaluation is not None
#                             else v.get('lazy_evaluation', True)})

#         data_type = type(data)
#         if data_type in (list, tuple):  # a sequence of dict, RLData, BaseRLData, etc.
#             self._data.extend(_from_dict(v) for v in data)  # each item in the sequence is dict-like
#         elif data_type in (BaseRLData, CategoricalData, NumericalData):
#             self._data.append(_from_dict(data))
#         elif data_type == dict:  # dict, RLData, BaseRLData, etc.
#             if isinstance(list(data.values())[0], dict):
#                 self._data.extend(_from_tuple(d) for d in data.items())
#             else:
#                 self._data.append(_from_dict(data))

#         elif isinstance(data, Generator):  # type: ignore
#             for v in data:
#                 if isinstance(v, dict):
#                     self._data.append(_from_dict(v))  # type: ignore
#                 elif isinstance(v, (list, tuple)):
#                     self._data.append(_from_tuple(v))  # type: ignore
#                 else:
#                     raise TypeError('Type is not recognized!')
#         else:
#             raise TypeError('Type is not recognized!')

#     @classmethod
#     def from_sparse_data(cls,
#                         value: Dict[str, T] = {},
#                         categorical: Dict[str, bool] = {},
#                         lower: Dict[str, Union[Number, None]] = {},
#                         upper: Dict[str, Union[Number, None]] = {},
#                         categories: Dict[str, Union[Categories, None]] = {},
#                         normalizer: Dict[str, Union[Normalizer, None]] = {},
#                         lazy_evaluation: Optional[bool] = None) -> RLData:
#         '''
#         Create an RLData instance.

#         Attributes:
#             value: value to store as a list or a dictionary
#             lower: minimum values for numerical components
#             upper: maximum values for numerical components
#             categories: set of categories for categorical components
#             categorical: whether each component is numerical (True) or categorical (False)
#             normalizer: a function that normalizes the respective component
#             lazy_evaluation: whether to store normalized values or compute on-demand (Default: False)
#         '''

#         temp = {}
#         for k, v in value.items():
#             temp[k] = dict(name=k,
#                            categorical=categorical[k],
#                            value=v,
#                            **{'categories': categories.get(k),
#                                'lower': lower.get(k),
#                                'upper': upper.get(k),
#                                'normalizer': normalizer.get(k), 'lazy_evaluation': lazy_evaluation})

#         return cls(temp)

#     def _clear_temps(self):
#         self._value = None
#         self._lower = None
#         self._upper = None
#         self._categories = None
#         self._categorical = None

#     def index(self, value: Any, start: int = 0, stop: Optional[int] = None) -> int:
#         _stop = stop if stop is not None else len(self._data)
#         if isinstance(value, BaseRLData):
#             for i in range(start, _stop):
#                 if self._data[i] == value:
#                     return i

#             raise ValueError(f'{value} is not on the list.')

#         elif isinstance(value, type(self._data[0].name)):
#             for i in range(start, _stop):
#                 if self._data[i].name == value:
#                     return i

#             raise ValueError(f'{value} is not on the list.')

#         else:
#             raise ValueError(f'{value} is not on the list.')

#     @property
#     def value(self) -> Dict[str, Any]:
#         '''Returns a dictionary of (name, RangedData) form.'''
#         if self._value is None:
#             self._value = dict((v.name, v.value) for v in self._data)

#         return self._value

#     @value.setter
#     def value(self, v: Dict[str, Any]):
#         self._value = None
#         for key, val in v.items():
#             self._data[self.index(key)].value = val

#     @property
#     def lower(self) -> Dict[str, Number]:
#         if self._lower is None:
#             self._lower = dict((v.name, v.lower) for v in self._data if hasattr(v, 'lower'))

#         return self._lower

#     @lower.setter
#     def lower(self, value: Dict[str, Number]) -> None:
#         self._lower = None
#         for key, val in value.items():
#             self._data[self.index(key)].lower = val

#     @property
#     def upper(self) -> Dict[str, Number]:
#         if self._upper is None:
#             self._upper = dict((v.name, v.upper) for v in self._data if hasattr(v, 'upper'))

#         return self._upper

#     @upper.setter
#     def upper(self, value: Dict[str, Number]) -> None:
#         self._upper = None
#         for key, val in value.items():
#             self._data[self.index(key)].upper = val

#     @property
#     def categories(self) -> Dict[str, Sequence[Any]]:
#         if self._categories is None:
#             self._categories = dict((v.name, v.categories) for v in self._data if hasattr(v, 'categories'))

#         return self._categories

#     @categories.setter
#     def categories(self, value: Dict[str, Sequence[Any]]) -> None:
#         self._categories = None
#         for key, val in value.items():
#             self._data[self.index(key)].categories = val

#     @property
#     def categorical(self) -> Dict[str, Sequence[bool]]:
#         if self._categorical is None:
#             self._categorical = dict((v.name, v.categorical) for v in self._data)

#         return self._categorical

#     @property
#     def normalized(self) -> RLData:
#         return RLData(tuple({'name': v.name, 'categorical': v.categorical,
#                        'value': v.normalized, 'lower': 0, 'upper': 1, 'lazy_evaluation': True} for v in self._data))

#     def flatten(self) -> List[Any]:
#         def make_iterable(x: Any) -> Iterable[Any]:
#             return x if isinstance(x, Iterable) else [x]

#         return list(itertools.chain(*[make_iterable(sublist)
#             for sublist in self.value.values()]))

#     def split(self) -> Union[RLData, List[RLData]]:  #, name_suffix: Optional[str] = None):
#         if len(self) == 1:
#             d = self._data[0]
#             if isinstance(d.value, (list, tuple)):
#                 val: Sequence[Any] = d.value
#                 name = d.name
#                 categorical = d.categorical
#                 lower = d.lower if not d.categorical else None
#                 upper = d.upper if not d.categorical else None
#                 categories = d.categories if d.categorical else None
#                 normalizer = d.normalizer
#                 lazy_evaluation = d.lazy_evaluation

#                 splitted_list = [RLData({
#                         'name': name,
#                         'value': v,
#                         'categorical': categorical,
#                         'lower': lower,
#                         'upper': upper,
#                         'categories': categories,
#                         'normalizer': normalizer,
#                         'lazy_evaluation': lazy_evaluation})
#                         for v in val]

#             else:
#                 splitted_list = RLData(self._data)

#         else:
#             splitted_list = [RLData({
#                     'name': d.name,
#                     'value': d.value,
#                     'categorical': d.categorical,
#                     'upper': d.upper if not d.categorical else None,
#                     'lower': d.lower if not d.categorical else None,
#                     'categories': d.categories if d.categorical else None,
#                     'normalizer': d.normalizer,
#                     'lazy_evaluation': d.lazy_evaluation})
#                     for d in self._data]

#         return splitted_list

#     def __getitem__(self, i: Union[int, slice]):
#         return self._data.__getitem__(i)

#     def __setitem__(self, i: Union[int, slice], o: Union[Any, Iterable[Any]]):
#         # TODO: I should somehow check the iterable to make sure it has proper data,
#         # but currently I have no idea how!
#         if not isinstance(o, (BaseRLData, Iterable)):
#             raise TypeError(
#                 'Only variables of type BaseRLData and subclasses are acceptable.')

#         self._clear_temps()
#         return self._data.__setitem__(i, o)

#     def __delitem__(self, i: Union[int, slice]) -> None:
#         self._data.__delitem__(i)

#     def insert(self, index: int, value: Any) -> None:
#         if not isinstance(value, BaseRLData):
#             raise TypeError(
#                 'Only variables of type BaseRLData and subclasses are acceptable.')

#         self._clear_temps()
#         self._data.insert(index, value)

#     def __len__(self) -> int:
#         return self._data.__len__()

#     def __hash__(self) -> int:
#         return hash(tuple(self._data))

#     def __eq__(self, other) -> bool:
#         return isinstance(other, type(self)) and (self._data == other._data)

#     def extend(self, values: R) -> None:
#         if isinstance(values, RLData):
#             for v in values:
#                 self._data.append(v)
#         else:
#             for v in values:
#                 self._data.extend(RLData(v))

#     def append(self, value: Union[Dict[str, Any], RLDataClass]) -> None:
#         if isinstance(value, RLData):
#             self._data.extend(value)
#         else:
#             self._data.extend(RLData(value))

#     def __add__(self, other: RLData) -> RLData:
#         if isinstance(other, BaseRLData):
#             if other.name in self._data:
#                 raise ValueError(
#                     'Cannot have items with same names. Use update() if you need to update an item.')
#         elif isinstance(other, RLData):
#             for k in other:
#                 if k in self._data:
#                     raise ValueError(
#                         'Cannot have items with same names. Use update() if you need to update an item.')
#         else:
#             raise TypeError(
#                 f'Concatenation of type RLData and {type(other)} not implemented!')

#                 # new_dict = other.as_dict()

#                 # new_dict = {v.name: v.as_dict() for v in other}

#         temp = copy.deepcopy(self)
#         temp.extend(other)
#         # temp.update(new_dict)

#         return temp

#     def __neg__(self) -> RLData:
#         temp = {v.name: v.as_dict() for v in self._data}
#         temp['value'] = -temp['value']

#         return RLData(temp)

#     def __repr__(self) -> str:
#         return f'[{super().__repr__()} -> {self._data}]'

#     def __str__(self) -> str:
#         return f"[{', '.join((d.__str__() for d in self._data))}]"


#     # # Mixin methods
#     # def append(self, value: _T) -> None: ...
#     # def clear(self) -> None: ...
#     # def extend(self, values: Iterable[_T]) -> None: ...
#     # def reverse(self) -> None: ...
#     # def pop(self, index: int = ...) -> _T: ...
#     # def remove(self, value: _T) -> None: ...
#     # def __iadd__(self, x: Iterable[_T]) -> MutableSequence[_T]: ...

#     # # Sequence Mixin methods
#     # def index(self, value: Any, start: int = ..., stop: int = ...) -> int: ...
#     # def count(self, value: Any) -> int: ...
#     # def __contains__(self, x: object) -> bool: ...
#     # def __iter__(self) -> Iterator[_T_co]: ...
#     # def __reversed__(self) -> Iterator[_T_co]: ...


# if __name__ == "__main__":
#     from timeit import timeit

#     def f1():
#         t1 = RLData.from_sparse_data({'test A': ['a', 'b']},
#                     categorical={'test A': True},
#                     categories={'test A': ['a', 'b']},
#                     # lower={'test B': 0},
#                     # upper={'test B': 100}
#                     )
#         return t1

#     def f2():
#         t2 = RLData(({'name': x, 'value': x, 'categories': list('abcdefghijklmnopqrstuvwxyz')}
#             for x in 'abcdefghijklmnopqrstuvwxyz'), lazy_evaluation=True)
#         return t2.normalized.flatten()

#     def f3():
#         t2 = RLData([{'name': 'A', 'value': 'a', 'categories': ['a', 'b']},
#             {'name': 'B', 'value': [10, 20], 'lower': 0, 'upper': 100}], lazy_evaluation=True)
#         return t2


#     # t1 = f1()
#     # t2 = RLData(t1)
#     print(f1() + f3())
#     # test = f1().split()
#     # print(test)
#     # print(timeit(f1, number=1000))
#     # print(timeit(f2, number=100))
#     # print(timeit(f3, number=1000))
#     # print(f2().normalized.as_list())
#     # print(a.values)
#     # print(a.lower)
#     # print(a.categories)
#     # a.values = {'test A':'b'}
#     # print(a + RLData({1: 100}, categorical={1: True}))
